#!/usr/bin/python
#dhn.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#

import os
import sys
import platform


def windows_and_frozen():
    if platform.uname()[0] != 'Windows':
        # return False
        # under Linux we have same problem too ...
        # I have no idea!
        return True

    #http://www.py2exe.org/index.cgi/HowToDetermineIfRunningFromExe
    import imp
    return  (hasattr(sys, "frozen") or # new py2exe
            hasattr(sys, "importers") or# old py2exe
            imp.is_frozen("__main__")) # tools/freeze



if __name__ == "__main__":
    # to be able to use "tail -f <log file>" command

#    OriginalSTDOUT = None
#    if platform.uname()[0] == 'Linux':
#        OriginalSTDOUT = sys.stdout
#        sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)

    # DHN use py2exe to build binaries under Windows, but it wont finish correctly!
    # After finishing all processes dhnmain.exe still working and eat 50% CPU! - it wont stop.
    # But if running from command line - it gives no problems.
    # It seems CSpace reactor makes some conflicts with twisted reactor.
    # However all threads are finished correctly - really strange thing.
    # TODO: I definately should find where is the problem here.
    # But let's take this for now - it is working.
    
    how_to_stop = windows_and_frozen()
    
    #---START THE MAIN CODE---
    import p2p.dhnmain
    ret = p2p.dhnmain.main()
    
#    if OriginalSTDOUT is not None:
#        sys.stdout = OriginalSTDOUT
    
    if how_to_stop:
        os._exit(ret)
    else:
        sys.exit(ret)




#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
import os
import sys
import errno
import time
import subprocess
import traceback

PIPE = subprocess.PIPE

#Pipe States:
PIPE_EMPTY = 0
PIPE_READY2READ = 1
PIPE_CLOSED = 2

if subprocess.mswindows:
    from win32file import ReadFile, WriteFile
    from win32pipe import PeekNamedPipe
    from win32api import TerminateProcess, OpenProcess, CloseHandle
    import msvcrt
else:
    import select
    import fcntl
    import signal

class Popen(subprocess.Popen):
    err_report = ''
    def __init__(self, args, bufsize=0, executable=None,
                stdin=None, stdout=None, stderr=None,
                preexec_fn=None, close_fds=False, shell=False,
                cwd=None, env=None, universal_newlines=False,
                startupinfo=None, creationflags=0):

        self.args = args
        subprocess.Popen.__init__(self,
            args, bufsize, executable,
            stdin, stdout, stderr,
            preexec_fn, close_fds, shell,
            cwd, env, universal_newlines,
            startupinfo, creationflags)

    def __del__(self):
        try:
            subprocess.Popen.__del__(self)
        except:
            pass

    def returncode(self):
        return self.returncode

    def recv(self, maxsize=None):
        r = self._recv('stdout', maxsize)
        if r is None:
            return ''
        return r

    def recv_err(self, maxsize=None):
        return self._recv('stderr', maxsize)

    def send_recv(self, input='', maxsize=None):
        return self.send(input), self.recv(maxsize), self.recv_err(maxsize)

    def get_conn_maxsize(self, which, maxsize):
        if maxsize is None:
            maxsize = 1024
        elif maxsize < 1:
            maxsize = 1
        return getattr(self, which), maxsize

    def _close(self, which):
        getattr(self, which).close()
        setattr(self, which, None)

    def state(self):
        return self._state('stdout')

    def make_nonblocking(self):
        if subprocess.mswindows:
            return
        conn, maxsize = self.get_conn_maxsize('stdout', None)
        if conn is None:
            return
        flags = fcntl.fcntl(conn, fcntl.F_GETFL)
        if not conn.closed:
            fcntl.fcntl(conn, fcntl.F_SETFL, flags| os.O_NONBLOCK)

    if subprocess.mswindows:
        def send(self, input):
            if not self.stdin:
                return None

            try:
                x = msvcrt.get_osfhandle(self.stdin.fileno())
                (errCode, written) = WriteFile(x, input)
            except:
                return None

            return written

        def _recv(self, which, maxsize):
            conn, maxsize = self.get_conn_maxsize(which, maxsize)
            if conn is None:
                return None

            try:
                x = msvcrt.get_osfhandle(conn.fileno())
                (read, nAvail, nMessage) = PeekNamedPipe(x, 0)
                if maxsize < nAvail:
                    nAvail = maxsize
                if nAvail > 0:
                    (errCode, read) = ReadFile(x, nAvail, None)
            except:
                return None

            if self.universal_newlines:
                read = self._translate_newlines(read)

            return read

        def _state(self, which):
            conn, maxsize = self.get_conn_maxsize(which, None)
            if conn is None:
                return PIPE_CLOSED
            try:
                x = msvcrt.get_osfhandle(conn.fileno())
            except:
                return PIPE_CLOSED
            try:
                (read, nAvail, nMessage) = PeekNamedPipe(x, 0)
            except:
                return PIPE_CLOSED
            if nAvail > 0:
                return PIPE_READY2READ
            return PIPE_EMPTY

        def kill(self):
            try:
                PROCESS_TERMINATE = 1
                handle = OpenProcess(PROCESS_TERMINATE, False, self.pid)
                TerminateProcess(handle, -1)
                CloseHandle(handle)
            except:
                pass


    else:
        def send(self, input):
            if not self.stdin:
                return None

            if not select.select([], [self.stdin], [], 0)[1]:
                return None

            try:
                written = os.write(self.stdin.fileno(), input)
            except:
                return None

            return written

        def _recv(self, which, maxsize):
            conn, maxsize = self.get_conn_maxsize(which, maxsize)
            if conn is None:
                return None

            flags = fcntl.fcntl(conn, fcntl.F_GETFL)
            if not conn.closed:
                fcntl.fcntl(conn, fcntl.F_SETFL, flags| os.O_NONBLOCK)

            try:
                if not select.select([conn], [], [], 0)[0]:
                    return None

                r = conn.read(maxsize)
                if not r:
                    return None

                if self.universal_newlines:
                    r = self._translate_newlines(r)
                return r
            finally:
                if not conn.closed:
                    fcntl.fcntl(conn, fcntl.F_SETFL, flags)

        def _state(self, which):
            conn, maxsize = self.get_conn_maxsize(which, None)
            if conn is None:
                return PIPE_CLOSED

            try:
                # check and see if there is any input ready
                ready = select.select([conn],[],[], 0)
                if conn in ready[0]:
                    return PIPE_READY2READ
                return PIPE_EMPTY
            except:
                return PIPE_CLOSED

        def kill(self):
            os.kill(self.pid, signal.SIGTERM)



def ExecuteString(execstr):
    try:
        import win32process
        return Popen(
            execstr,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            creationflags = win32process.CREATE_NO_WINDOW,)
    except:
        return None



from OpenSSL import SSL
from OpenSSL.crypto import load_privatekey, load_certificate
from OpenSSL.crypto import FILETYPE_ASN1, FILETYPE_PEM
#from OpenSSL.SSL import SSLv2_METHOD, SSLv3_METHOD, SSLv23_METHOD, TLSv1_METHOD
#from OpenSSL.SSL import VERIFY_PEER, VERIFY_FAIL_IF_NO_PEER_CERT, VERIFY_CLIENT_ONCE, VERIFY_NONE
from OpenSSL.SSL import OP_NO_SSLv2 , OP_NO_SSLv3 , OP_SINGLE_DH_USE

SSL_METHOD_SSLv2 = 0
SSL_METHOD_SSLv3 = 1
SSL_METHOD_TLSv1 = 2
SSL_METHOD_SSLv23 = 3
SSL_METHOD_MAX = 4

SSL_METHOD_TYPE_CLIENT = 0
SSL_METHOD_TYPE_SERVER = 1
SSL_METHOD_TYPE_GENERIC = 2
SSL_METHOD_TYPE_MAX = 3

SSL_VERIFY_MODE_NONE = 0
SSL_VERIFY_MODE_SELF_SIGNED = 1
SSL_VERIFY_MODE_MAX = 2

SSL_OP_NO_SSLv2 = SSL.OP_NO_SSLv2
SSL_OP_NO_SSLv3 = SSL.OP_NO_SSLv3
SSL_OP_SINGLE_DH_USE = SSL.OP_SINGLE_DH_USE

class SSLError(Exception):
    def __init__( self, *args ) :
        Exception.__init__( self, *args )

class SSLWantError( SSLError ): 
    pass

class SSLWantReadError( SSLWantError ):
    pass

class SSLWantWriteError( SSLWantError ):
    pass

class SSLWantX509LookupError( SSLWantError ):
    pass

class SSLContext():
    def __init__( self, sslMethod ) :
        m = {SSL_METHOD_SSLv2: SSL.SSLv2_METHOD,
             SSL_METHOD_SSLv3: SSL.SSLv3_METHOD,
             SSL_METHOD_TLSv1: SSL.TLSv1_METHOD,
             SSL_METHOD_SSLv23: SSL.SSLv23_METHOD,}.get(sslMethod, SSL.TLSv1_METHOD)
        self.c = SSL.Context(m)
        self.verifyCallback = None
        
    def setCertificate(self, cert) :
        self.c.use_certificate(cert)
        
    def setPrivateKey(self, rsakey) :
        self.c.use_privatekey(load_privatekey(FILETYPE_ASN1, rsakey.toDER_PrivateKey()))
        
    def checkPrivateKey(self):
        self.c.check_privatekey()
        
    def setVerifyMode(self, m, callback=None):
        cb = self.selfSignedVerifyCallback
        if callback is not None:
            cb = callback
        if m == SSL_VERIFY_MODE_NONE:
            self.c.set_verify(SSL.VERIFY_NONE, cb)
        elif m == SSL_VERIFY_MODE_SELF_SIGNED:
            self.c.set_verify(SSL.VERIFY_PEER|SSL.VERIFY_FAIL_IF_NO_PEER_CERT, cb)
    
    def setOptions(self, flags):
        return self.c.set_options(flags)
    
    def selfSignedVerifyCallback(self, conn, cert, errno, depth, preverify_ok):
        return preverify_ok
    
    def loadVerifyLocations(self, path):
        self.c.load_verify_locations(path)
    
    
class SSLConnection():
    def __init__( self, sslContext, sock ) :
        self.conn = SSL.Connection(sslContext.c, sock)
        self.sock = sock
        
    def setConnectState(self):
        self.conn.set_connect_state()
        
    def setAcceptState(self):
        self.conn.set_accept_state()
        
    def connect(self):
        try:
            self.conn.do_handshake()
        except SSL.WantReadError:
            raise SSLWantReadError
        except SSL.WantWriteError:
            raise SSLWantWriteError
        except SSL.WantX509LookupError:
            raise SSLWantX509LookupError
        except SSL.Error, e:
            raise SSLError, e
        except:
            raise

    def accept(self):
        try:
            self.conn.do_handshake()
        except SSL.WantReadError:
            raise SSLWantReadError
        except SSL.WantWriteError:
            raise SSLWantWriteError
        except SSL.WantX509LookupError:
            raise SSLWantX509LookupError
        except SSL.Error, e:
            raise SSLError, e
        except:
            raise
        
    def getPeerCertificate(self):
        return  self.conn.get_peer_certificate()
    
    def pending(self):
        return self.conn.pending()
    
    def recv(self, bufsize):
        try:
            return self.conn.recv(bufsize)
        except SSL.WantReadError:           
            raise SSLWantReadError
        except SSL.WantWriteError:          
            raise SSLWantWriteError
        except SSL.WantX509LookupError:     
            raise SSLWantX509LookupError
        except SSL.Error, e:                   
            raise SSLError, e 
        except:                             
            raise
    
    def send(self, data):
        try:
            return self.conn.send(data)    
        except SSL.WantReadError:
            raise SSLWantReadError
        except SSL.WantWriteError:
            raise SSLWantWriteError
        except SSL.WantX509LookupError:
            raise SSLWantX509LookupError
        except SSL.Error, e:
            raise SSLError, e  
        except:
            raise

    def shutdown(self):
        self.sock = None
        try:
            return self.conn.shutdown()
        except SSL.WantReadError:
            raise SSLWantReadError
        except SSL.WantWriteError:
            raise SSLWantWriteError
        except SSL.WantX509LookupError:
            raise SSLWantX509LookupError
        except SSL.Error, e:
            raise SSLError, e 
        except:
            raise




import logging
from time import time
from socket import error as sock_error
from ncrypt0.rand import bytes as rand_bytes
from nitro.tcp import TCPStream
from nitro.tcpbridge import TCPBridge
from cspace.util.rpc import RPCConnection

logger = logging.getLogger('cspace.network.router')

ROUTER_IDLE_TIMEOUT = 30*60

class RouterState( object ) :
    def __init__( self ) :
        self.handlers = {}
        self.registered = {}
        self.connecting = {}
        self.bridgeInitiators = {}
        self.bridges = {}

def startBridge( sock1, sock2, reactor, routerState ) :
    bridge = TCPBridge( sock1, sock2, reactor )
    routerState.bridges[bridge] = 1
    def onClose() :
        logger.info('delete bridge %s' % bridge.id)
        del routerState.bridges[bridge]
    bridge.setCloseCallback( onClose )

class BridgeInitiator( object ) :
    def __init__( self, sock1, pendingWrite1, sock2, pendingWrite2, reactor, routerState ) :
        self.reactor = reactor
        self.streams = []
        self._initStream( sock1, pendingWrite1 )
        self._initStream( sock2, pendingWrite2 )
        self.written = 0
        self.routerState = routerState
        self.routerState.bridgeInitiators[self] = 1

    def _initStream( self, sock, pendingWrite ) :
        assert pendingWrite
        stream = TCPStream( sock, self.reactor )
        self.streams.append( stream )
        stream.setCloseCallback( self.close )
        stream.setErrorCallback( lambda e,m : self.close() )
        stream.setWriteCompleteCallback( self._onWritten )
        stream.writeData( pendingWrite )

    def close( self ) :
        for s in self.streams : s.close()
        del self.stream[:]
        del self.routerState.bridgeInitiators[self]

    def _onWritten( self ) :
        self.written += 1
        if self.written < 2 : return
        socks = [s.getSock() for s in self.streams]
        for s in self.streams : s.shutdown()
        del self.streams[:]
        del self.routerState.bridgeInitiators[self]
        startBridge( socks[0], socks[1], self.reactor, self.routerState )

class RouterServerHandler( object ) :
    (INITIAL,REGISTERED,CONNECTING,CLOSED) = range(4)
    def __init__( self, sock, reactor, routerState ) :
        self.rpcConn = RPCConnection( sock, reactor )
        self.reactor = reactor
        self.routerState = routerState
        self.rpcConn.setCloseCallback( self.close )
        self.rpcConn.setRequestCallback( self._onRequest )
        self.requestTable = {}
        for msg in 'Ping Register Connect Accept'.split() :
            self.requestTable[msg] = getattr( self, '_do%s' % msg )
        self.routerState.handlers[self] = 1
        self.state = self.INITIAL
        self._lastActiveTime = time()

    def close( self ) :
        self.rpcConn.close()
        del self.routerState.handlers[self]
        if self.state == self.REGISTERED :
            del self.routerState.registered[self.routerId]
        elif self.state == self.CONNECTING :
            del self.routerState.connecting[self.connectionId]
        self.state = self.CLOSED

    def lastActiveTime( self ) :
        return self._lastActiveTime

    def sendIncoming( self, connectionId ) :
        assert self.state == self.REGISTERED
        self.rpcConn.oneway( connectionId )

    def _onRequest( self, payload, ctx ) :
        self._lastActiveTime = time()
        try :
            assert type(payload) is list
            assert len(payload) >= 1
            msg = payload.pop( 0 )
            assert type(msg) is str
            handler = self.requestTable.get( msg )
            assert handler is not None
        except :
            ctx.response( [-1] )
            return
        handler( payload, ctx )

    def _doPing( self, args, ctx ) :
        try :
            peerAddr = self.rpcConn.getSock().getpeername()
        except sock_error :
            ctx.response( [-1] )
            return
        ctx.response( (0,peerAddr) )

    def _doRegister( self, args, ctx ) :
        try :
            assert self.state == self.INITIAL
            assert len(args) == 0
        except :
            ctx.response( (-1,'') )
            return
        while True :
            routerId = rand_bytes( 20 )
            if routerId not in self.routerState.registered : break
        self.routerId = routerId
        self.state = self.REGISTERED
        self.routerState.registered[routerId] = self
        ctx.response( (0,self.routerId) )

    def _doConnect( self, args, ctx ) :
        try :
            assert self.state == self.INITIAL
            assert len(args) == 1
            routerId = args[0]
            assert type(routerId) is str
            handler = self.routerState.registered.get( routerId )
            assert handler is not None
        except :
            ctx.response( [-1] )
            return
        while True :
            connectionId = rand_bytes( 20 )
            if connectionId not in self.routerState.connecting : break
        self.connectionId = connectionId
        self.connectCtx = ctx
        self.state = self.CONNECTING
        self.routerState.connecting[connectionId] = self
        handler.sendIncoming( connectionId )

    def finishConnect( self ) :
        assert self.state == self.CONNECTING
        self.connectCtx.response( [0] )
        del self.routerState.connecting[self.connectionId]
        del self.routerState.handlers[self]
        sock = self.rpcConn.getSock()
        pendingWrite = self.rpcConn.getPendingWrite()
        self.rpcConn.shutdown()
        self.state = self.CLOSED
        return (sock,pendingWrite)

    def _doAccept( self, args, ctx ) :
        try :
            assert self.state == self.INITIAL
            assert len(args) == 1
            connectionId = args[0]
            assert type(connectionId) is str
            handler = self.routerState.connecting.get( connectionId )
            assert handler is not None
        except :
            ctx.response( [-1] )
            return
        sock1,pendingWrite1 = handler.finishConnect()
        ctx.response( [0] )
        del self.routerState.handlers[self]
        sock2 = self.rpcConn.getSock()
        pendingWrite2 = self.rpcConn.getPendingWrite()
        self.rpcConn.shutdown()
        self.state = self.CLOSED
        BridgeInitiator( sock1, pendingWrite1, sock2, pendingWrite2, self.reactor, self.routerState )

class Router( object ) :
    def __init__( self, listener, reactor ) :
        self.listener = listener
        self.reactor = reactor
        self.listener.setCallback( self._onIncoming )
        self.listener.enable( True )
        self.routerState = RouterState()
        self.timerOp = reactor.addTimer( ROUTER_IDLE_TIMEOUT/2, self._onTimer )

    def close( self ) :
        for handler in self.routerState.handlers.keys() :
            handler.close()
        for initiator in self.routerState.bridgeInitiators.keys() :
            initiator.close()
        for bridge in self.routerState.bridges.keys() :
            bridge.close()
        self.routerState.bridges.clear()
        self.listener.close()
        self.listener = None
        self.timerOp.cancel()
        self.timerOp = None

    def _onIncoming( self, sock ) :
        global logger
        logger.info('on incoming from ' + str(sock.getpeername()))
        RouterServerHandler( sock, self.reactor, self.routerState )

    def _onTimer( self ) :
        curTime = time()
        for bridge in self.routerState.bridges.keys() :
            idleTime = curTime - bridge.lastActiveTime()
            if idleTime > ROUTER_IDLE_TIMEOUT :
                bridge.close()
                del self.routerState.bridges[bridge]
        for handler in self.routerState.handlers.keys() :
            idleTime = curTime - handler.lastActiveTime()
            if idleTime > ROUTER_IDLE_TIMEOUT :
                handler.close()



import os, re, StringIO
from cspace.main.common import profileSettings

permissionsHelp ="""\
For EVERY connection, rules are matched in SEQUENTIAL order.
If two rules match a particular connection, then the first
rule will take precedence

Format:
    (action-spec) (user-spec) service (service-spec)
Where:
    action-spec  -> allow | deny | prompt
    user-spec    -> nickname | <contact-user> | <any-user>
    service-spec -> servicename | <any-service>

Example:
allow  sreeram        service TextChat
allow  <contact-user> service FileTransfer
deny   <any-user>     service FileTransfer
prompt <any-user>     service <any-service>
"""

predefinedPermissions ="""\
prompt <contact-user> service RemoteDesktop
allow  <contact-user> service <any-service>
deny   <any-user>     service RemoteDesktop
allow  <any-user>     service TextChat
prompt <any-user>     service <any-service>
"""

ACCESS_ALLOW = 0
ACCESS_DENY = 1
ACCESS_PROMPT = 2
ACCESS_COUNT = 3

def _loadPredefinedPermissions() :
    return predefinedPermissions

def _loadUserPermissions( profile ) :
    entry = profile.storeEntry + '/Permissions'
    return profileSettings().getData( entry, '' )

def _storeUserPermissions( profile, data ) :
    entry = profile.storeEntry + '/Permissions'
    return profileSettings().setData( entry, data )

class BaseUser( object ) :
    def matches( self, userName ) :
        raise NotImplementedError
    def toString( self ) :
        raise NotImplementedError

class NamedUser( BaseUser ) :
    def __init__( self, userName ) :
        self.userName = userName
    def matches( self, userName ) :
        return userName == self.userName
    def toString( self ) :
        return self.userName

class ContactUser( BaseUser ) :
    def __init__( self, profile ) :
        self.profile = profile
    def matches( self, userName ) :
        contact = self.profile.getContactByName( userName )
        if contact is None : return False
        return True
    def toString( self ) :
        return '<contact-user>'

class AnyUser( BaseUser ) :
    def __init__( self ) :
        pass
    def matches( self, userName ) :
        return True
    def toString( self ) :
        return '<any-user>'

class BaseService( object ) :
    def matches( self, serviceName ) :
        raise NotImplementedError
    def toString( self ) :
        raise NotImplementedError

class NamedService( BaseService ) :
    def __init__( self, serviceName ) :
        self.serviceName = serviceName
    def matches( self, serviceName ) :
        return serviceName == self.serviceName
    def toString( self ) :
        return self.serviceName

class AnyService( BaseService ) :
    def __init__( self ) :
        pass
    def matches( self, serviceName ) :
        return True
    def toString( self ) :
        return '<any-service>'

class BadRuleError( Exception ) :
    def __init__( self, lineNumber=0 ) :
        Exception.__init__( self )
        self.lineNumber = lineNumber
    def setLineNumber( self, lineNumber ) :
        self.lineNumber = lineNumber
class BadUserInRuleError( BadRuleError ) :
    def __init__( self, user, lineNumber=0 ) :
        BadRuleError.__init__( self, lineNumber )
        self.user = user
class BadServiceInRuleError( BadRuleError ) :
    def __init__( self, service, lineNumber=0 ) :
        BadRuleError.__init__( self, lineNumber )
        self.service = service

class BlankRule( object ) :
    def __init__( self, line ) :
        self.line = line

    def execute( self, userName, serviceName ) :
        return None

    def toLine( self ) :
        return self.line

class PermissionRule( object ) :
    def __init__( self, accessAction, userMatcher, serviceMatcher ) :
        self.s = [''] * 5
        self.s[0] = ''
        self.accessAction = accessAction
        self.s[1] = ''
        self.userMatcher = userMatcher
        self.s[2] = ''
        self.s[3] = ''
        self.serviceMatcher = serviceMatcher
        self.s[4] = ''

    def execute( self, userName, serviceName ) :
        if (self.userMatcher.matches(userName) and
                self.serviceMatcher.matches(serviceName)) :
            return self.accessAction
        return None

    def toLine( self ) :
        actions = [ 'allow', 'deny', 'prompt' ]
        act = actions[ self.accessAction ]
        user = self.userMatcher.toString()
        service = self.serviceMatcher.toString()
        s = self.s
        return ''.join( [s[0],act,s[1],user,s[2],'service',s[3],service,s[4]] )

    def fromLine( line, profile, serviceList ) :
        comment = ''
        x = line.find('#')
        if x >= 0 :
            line,comment = line[:x],line[x:]
        if not line.strip() :
            return BlankRule( line+comment )
        toks = re.split( '(\S+)', line )
        if len(toks) != 9 :
            raise BadRuleError()
        s = [''] * 5
        s[0],act,s[1],user,s[2],dummy,s[3],service,s[4] = toks
        s[4] += comment
        if dummy != 'service' :
            raise BadRuleError()
        actions = dict( allow=ACCESS_ALLOW, deny=ACCESS_DENY,
                prompt=ACCESS_PROMPT )
        accessAction = actions.get( act.lower(), None )
        if accessAction is None :
            raise BadRuleError()
        if user.lower() == '<contact-user>' :
            userMatcher = ContactUser( profile )
        elif user.lower() == '<any-user>' :
            userMatcher = AnyUser()
        else :
            contact = profile.getContactByName( user )
            if contact is None :
                raise BadUserInRuleError( user )
            userMatcher = NamedUser( user )
        if service.lower() == '<any-service>' :
            serviceMatcher = AnyService()
        else :
            if service not in serviceList :
                raise BadServiceInRuleError( service )
            serviceMatcher = NamedService( service )
        rule = PermissionRule( accessAction, userMatcher, serviceMatcher )
        rule.s = s
        return rule
    fromLine = staticmethod(fromLine)

class Permissions( object ) :
    def __init__( self, profile, serviceList ) :
        self.profile = profile
        self.serviceList = serviceList
        self.predefinedRules = []
        self.userRules = []
        self.modified = False

        data = _loadUserPermissions( profile )
        self.setUserPermissions( data, ignoreBadRules=True )
        data = _loadPredefinedPermissions()
        self.setPredefinedPermissions( data, ignoreBadRules=True )

    def isModified( self ) : return self.modified
    def setModified( self, modified ) :
        self.modified = modified

    def execute( self, userName, serviceName ) :
        rules = self.userRules + self.predefinedRules
        for r in rules :
            result = r.execute( userName, serviceName )
            if result is not None : return result
        return ACCESS_PROMPT

    def _parsePermissions( self, data, ignoreBadRules=False ) :
        rules = []
        hasBadRules = False
        for num,line in enumerate(data.splitlines()) :
            try :
                rule = PermissionRule.fromLine( line, self.profile,
                        self.serviceList )
                rules.append( rule )
            except BadRuleError, e :
                if not ignoreBadRules :
                    e.setLineNumber( num+1 )
                    raise
                line = '# BadRule: ' + line
                rule = BlankRule( line )
                rules.append( rule )
                hasBadRules = True
        return rules,hasBadRules

    def setPredefinedPermissions( self, data, ignoreBadRules=False ) :
        rules,hasBadRules = self._parsePermissions( data,
                ignoreBadRules )
        self.predefinedRules = rules

    def setUserPermissions( self, data, ignoreBadRules=False ) :
        rules,hasBadRules = self._parsePermissions( data,
                ignoreBadRules )
        self.userRules = rules
        self.modified = hasBadRules

    def getHelpText( self ) :
        return permissionsHelp

    def getPredefinedPermissions( self ) :
        out = StringIO.StringIO()
        for rule in self.predefinedRules :
            print>>out, rule.toLine()
        return out.getvalue()

    def getUserPermissions( self ) :
        out = StringIO.StringIO()
        for rule in self.userRules :
            print>>out, rule.toLine()
        return out.getvalue()

    def changeContactName( self, oldName, newName ) :
        for rule in self.userRules :
            if isinstance(rule,PermissionRule) :
                userMatcher = rule.userMatcher
                if isinstance(userMatcher,NamedUser) :
                    if oldName == userMatcher.userName :
                        userMatcher.userName = newName
                        self.modified = True

    def removeContact( self, name ) :
        newRules = []
        for rule in self.userRules :
            if isinstance(rule,PermissionRule) and \
                    isinstance(rule.userMatcher,NamedUser) and \
                    (rule.userMatcher.userName == name) :
                rule = BlankRule( '# Contact removed: ' + rule.toLine() )
                self.modified = True
            newRules.append( rule )
        self.userRules = newRules

    def savePermissions( self ) :
        perm = self.getUserPermissions()
        _storeUserPermissions( self.profile, perm )
        self.modified = False



#!/usr/bin/python
#block_rebuilder.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#

import os
import sys
import time
import random


try:
    from twisted.internet import reactor
except:
    sys.exit('Error initializing twisted.internet.reactor in block_rebuilder.py')

from twisted.internet.defer import Deferred, maybeDeferred
from twisted.internet import threads


import lib.dhnio as dhnio
import lib.misc as misc
import lib.nameurl as nameurl
import lib.transport_control as transport_control
import lib.settings as settings
import lib.contacts as contacts
import lib.eccmap as eccmap
import lib.tmpfile as tmpfile
import lib.diskspace as diskspace
import lib.packetid as packetid

import fire_hire
import backup_rebuilder
import list_files_orator 

import backup_matrix
import raidread
import p2p_service
import io_throttle
import contact_status
import data_sender

#------------------------------------------------------------------------------ 

class BlockRebuilder():
    def __init__(self,  
                 eccMap, 
                 backupID, 
                 blockNum, 
                 supplierSet, 
                 remoteData, 
                 remoteParity,
                 localData, 
                 localParity, 
                 creatorId = None, 
                 ownerId = None):
        self.eccMap = eccMap
        self.backupID = backupID
        self.blockNum = blockNum
        self.supplierSet = supplierSet
        self.supplierCount = len(self.supplierSet.suppliers)
        self.remoteData = remoteData
        self.remoteParity = remoteParity
        self.localData = localData
        self.localParity = localParity
        self.creatorId = creatorId
        self.ownerId = ownerId
        # at some point we may be dealing with when we're scrubbers
        if self.creatorId == None:
            self.creatorId = misc.getLocalID()
        if self.ownerId == None:
            self.ownerId = misc.getLocalID()
        # this files we want to rebuild
        # need to identify which files to work on
        self.missingData = [0] * self.supplierCount
        self.missingParity = [0] * self.supplierCount
        # array to remember requested files
        self.reconstructedData = [0] * self.supplierCount
        self.reconstructedParity = [0] * self.supplierCount

    def IdentifyMissing(self):
        self.availableSuppliers = self.supplierSet.GetActiveArray()
        for supplierNum in xrange(self.supplierCount):
            if self.availableSuppliers[supplierNum] == 0:
                continue
            # if remote Data file not exist and supplier is online
            # we mark it as missing and will try to rebuild this file and send to him
            if self.remoteData[supplierNum] != 1:
                # mark file as missing  
                self.missingData[supplierNum] = 1
            # same for Parity file
            if self.remoteParity[supplierNum] != 1:
                self.missingParity[supplierNum] = 1

    def IsMissingFilesOnHand(self):
        for supplierNum in xrange(self.supplierCount):
            # if supplier do not have the Data but is on line 
            if self.missingData[supplierNum] == 1:
                # ... and we also do not have the Data 
                if self.localData[supplierNum] != 1:
                    # return False - will need request the file   
                    return False
            # same for Parity                
            if self.missingParity[supplierNum] == 1:
                if self.localParity[supplierNum] != 1:
                    return False
        return True

    def BuildRaidFileName(self, supplierNumber, dataOrParity):
        return os.path.join(settings.getLocalBackupsDir(), self.BuildFileName(supplierNumber, dataOrParity))

    def BuildFileName(self, supplierNumber, dataOrParity):
        return packetid.MakePacketID(self.backupID, self.blockNum, supplierNumber, dataOrParity)

    def HaveAllData(self, parityMap):
        for segment in parityMap:
            if self.localData[segment] == 0:
                return False
        return True

    def AttemptRebuild(self):
        dhnio.Dprint(10, 'block_rebuilder.AttemptRebuild BEGIN')
        newData = False
        madeProgress = True
        while madeProgress:
            madeProgress = False
            # if number of suppliers were changed - stop immediately 
            if contacts.numSuppliers() != self.supplierCount:
                dhnio.Dprint(10, 'block_rebuilder.AttemptRebuild END - number of suppliers were changed')
                return False
            # will check all data packets we have 
            for supplierNum in xrange(self.supplierCount):
                dataFileName = self.BuildRaidFileName(supplierNum, 'Data')
                # if we do not have this item on hands - we will reconstruct it from other items 
                if self.localData[supplierNum] == 0:
                    parityNum, parityMap = self.eccMap.GetDataFixPath(self.localData, self.localParity, supplierNum)
                    if parityNum != -1:
                        rebuildFileList = []
                        rebuildFileList.append(self.BuildRaidFileName(parityNum, 'Parity'))
                        for supplierParity in parityMap:
                            if supplierParity != supplierNum:
                                rebuildFileList.append(self.BuildRaidFileName(supplierParity, 'Data'))
                        dhnio.Dprint(10, '    rebuilding file %s from %d files' % (os.path.basename(dataFileName), len(rebuildFileList)))
                        raidread.RebuildOne(rebuildFileList, len(rebuildFileList), dataFileName)
                    if os.path.exists(dataFileName):
                        self.localData[supplierNum] = 1
                        madeProgress = True
                        dhnio.Dprint(10, '        Data file %s found after rebuilding for supplier %d' % (os.path.basename(dataFileName), supplierNum))
                # now we check again if we have the data on hand after rebuild at it is missing - send it
                # but also check to not duplicate sending to this man   
                # now sending is separated, see the file data_sender.py          
                if self.localData[supplierNum] == 1 and self.missingData[supplierNum] == 1: # and self.dataSent[supplierNum] == 0:
                    dhnio.Dprint(10, '            rebuilt a new Data for supplier %d' % supplierNum)
                    newData = True
                    self.reconstructedData[supplierNum] = 1
                    # self.outstandingFilesList.append((dataFileName, self.BuildFileName(supplierNum, 'Data'), supplierNum))
                    # self.dataSent[supplierNum] = 1
        # now with parities ...            
        for supplierNum in xrange(self.supplierCount):
            parityFileName = self.BuildRaidFileName(supplierNum, 'Parity')
            if self.localParity[supplierNum] == 0:
                parityMap = self.eccMap.ParityToData[supplierNum]
                if self.HaveAllData(parityMap):
                    rebuildFileList = []
                    for supplierParity in parityMap:
                        rebuildFileList.append(self.BuildRaidFileName(supplierParity, 'Data')) # ??? why not 'Parity'
                    dhnio.Dprint(10, '    rebuilding file %s from %d files' % (os.path.basename(parityFileName), len(rebuildFileList)))
                    raidread.RebuildOne(rebuildFileList, len(rebuildFileList), parityFileName)
                    if os.path.exists(parityFileName):
                        dhnio.Dprint(10, '        Parity file %s found after rebuilding for supplier %d' % (os.path.basename(parityFileName), supplierNum))
                        self.localParity[supplierNum] = 1
            # so we have the parity on hand and it is missing - send it
            if self.localParity[supplierNum] == 1 and self.missingParity[supplierNum] == 1: # and self.paritySent[supplierNum] == 0:
                dhnio.Dprint(10, '            rebuilt a new Parity for supplier %d' % supplierNum)
                newData = True
                self.reconstructedParity[supplierNum] = 1
                # self.outstandingFilesList.append((parityFileName, self.BuildFileName(supplierNum, 'Parity'), supplierNum))
                # self.paritySent[supplierNum] = 1
        dhnio.Dprint(10, 'block_rebuilder.AttemptRebuild END')
        return newData

    def WorkDoneReport(self):
        for supplierNum in xrange(self.supplierCount):
            if self.localData[supplierNum] == 1 and self.reconstructedData[supplierNum] == 1:
                backup_matrix.LocalFileReport(None, self.backupID, self.blockNum, supplierNum, 'Data')
            if self.localParity[supplierNum] == 1 and self.reconstructedParity[supplierNum] == 1:
                backup_matrix.LocalFileReport(None, self.backupID, self.blockNum, supplierNum, 'Parity')
                    
                    


#!/usr/bin/python
#dhnview.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#

if __name__ == "__main__":
    import p2p.dhnview
    p2p.dhnview.main()





#!/usr/bin/env python
#shutdowner.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#

import os
import sys

try:
    from twisted.internet import reactor
except:
    sys.exit('Error initializing twisted.internet.reactor in shutdowner.py')
from twisted.internet.defer import Deferred, maybeDeferred
from twisted.internet.task import LoopingCall

import lib.dhnio as dhnio
from lib.automat import Automat
import lib.automats as automats

import initializer
import dhninit

_Shutdowner = None

#------------------------------------------------------------------------------

def A(event=None, arg=None):
    global _Shutdowner
    if _Shutdowner is None:
        _Shutdowner = Shutdowner('shutdowner', 'AT_STARTUP', 2)
    if event is not None:
        _Shutdowner.event(event, arg)
    return _Shutdowner


class Shutdowner(Automat):
    
    def init(self):
        self.flagApp = False
        self.flagReactor = False
        self.shutdown_param = None
    
    def state_changed(self, oldstate, newstate):
        automats.set_global_state('SHUTDOWN ' + newstate)
        initializer.A('shutdowner.state', newstate)

    def A(self, event, arg):
        #---AT_STARTUP---
        if self.state is 'AT_STARTUP':
            if event == 'init' :
                self.state = 'INIT'
                self.flagApp=False
                self.flagReactor=False
        #---INIT---
        elif self.state is 'INIT':
            if event == 'stop' :
                self.doSaveParam(arg)
                self.flagApp=True
            elif event == 'reactor-stopped' :
                self.flagReactor=True
            elif event == 'ready' and self.flagReactor :
                self.state = 'FINISHED'
            elif event == 'ready' and not self.flagReactor and self.flagApp :
                self.state = 'STOPPING'
                self.doShutdown(arg)
            elif event == 'ready' and not self.flagReactor and not self.flagApp :
                self.state = 'READY'
        #---READY---
        elif self.state is 'READY':
            if event == 'stop' :
                self.state = 'STOPPING'
                self.doShutdown(arg)
            elif event == 'reactor-stopped' :
                self.state = 'FINISHED'
            elif event == 'block' :
                self.state = 'BLOCKED'
        #---BLOCKED---
        elif self.state is 'BLOCKED':
            if event == 'stop' :
                self.doSaveParam(arg)
                self.flagApp=True
            elif event == 'reactor-stopped' :
                self.flagReactor=True
            elif event == 'unblock' and not self.flagReactor and not self.flagApp :
                self.state = 'READY'
            elif event == 'unblock' and not self.flagReactor and self.flagApp :
                self.state = 'STOPPING'
                self.doShutdown(arg)
            elif event == 'unblock' and self.flagReactor :
                self.state = 'FINISHED'
        #---FINISHED---
        elif self.state is 'FINISHED':
            pass
        #---STOPPING---
        elif self.state is 'STOPPING':
            if event == 'reactor-stopped' :
                self.state = 'FINISHED'

    def doSaveParam(self, arg):
        self.shutdown_param = arg
        dhnio.Dprint(2, 'shutdowner.doSaveParam %s' % str(self.shutdown_param))

    def doShutdown(self, arg):
        param = arg
        if self.shutdown_param is not None:
            param = self.shutdown_param
        if arg is None:
            param = 'exit' 
        elif isinstance(arg, str):
            param = arg
        if param not in ['exit', 'restart', 'restartnshow']:
            param = 'exit'
        if param == 'exit':
            dhninit.shutdown_exit()
        elif param == 'restart':
            dhninit.shutdown_restart()
        elif param == 'restartnshow':
            dhninit.shutdown_restart('show')









class AsyncOp( object ) :
    def __init__( self, callback=None, canceler=None ) :
        self.callback = callback
        self.canceler = canceler

    def setCallback( self, callback ) :
        self.callback = callback
        return self

    def setCanceler( self, canceler ) :
        self.canceler = canceler
        return self

    def cancel( self ) :
        if self.canceler : self.canceler()

    def notify( self, *args, **kwargs ) :
        if self.callback : self.callback( *args, **kwargs )

Op = AsyncOp



#!/usr/bin/env python
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#

import os
import sys
from twisted.internet.defer import Deferred
from twisted.python import log
from twisted.python import components
from twisted.cred import portal
from twisted.conch.client import options
from twisted.conch.client import direct
from twisted.conch.ssh import keys
from twisted.conch.ssh import session

try:
    from twisted.internet import reactor
except:
    sys.exit('Error initializing twisted.internet.reactor in transport_ssh.py')

if __name__ == '__main__':
    dirpath = os.path.dirname(os.path.abspath(sys.argv[0]))
    sys.path.insert(0, os.path.abspath('datahaven'))
    sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..')))
    sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..', '..')))

import misc
import settings
import dhnio
import dhncrypto
import contacts
import identitycache
import transport_control


from transport_ssh_client import MySSHConnection
from transport_ssh_client import MyUserAuth
from transport_ssh_client import my_verifyHostKey

from transport_ssh_server import MyAvatar
from transport_ssh_server import MyRealm
from transport_ssh_server import MyPublicKeyChecker
from transport_ssh_server import MyFactory
from transport_ssh_server import pub_keys_db


##import datahaven.lib.settings as settings

##log.startLogging(sys.stderr)

#-------------------------------------------------------------------------------
# PUBLIC KEYS DATABASE
#-------------------------------------------------------------------------------

def add_key(name, public_key):
    global pub_keys_db
    if pub_keys_db.has_key(name):
        return
    pub_keys_db[name] = keys.Key.fromString( public_key )
    transport_control.log('ssh', 'add public key to db for ' + name)
    dhnio.Dprint(10, "transport_ssh.add_key for " + name)

def remove_key(name):
    global pub_keys_db
    transport_control.log('ssh', 'remove public key of ' + name)
    if (pub_keys_db.has_key(name)):
        pub_keys_db.pop(name)

def add_identity2db(idlist):
    def _add_ok(src, idurl):
#        words = src.split('\n', 1)
#        if len(words) < 1:
#            dhnio.Dprint(1, 'transport_ssh.add_key ERROR wrong data')
#            return
#        ident = identitycache.FromCache(words[0].strip())
        ident = identitycache.FromCache(idurl)
        if ident == '' or ident is None:
            dhnio.Dprint(1, 'transport_ssh.add_key ERROR still not in the cache')
            return
        add_key(ident.getIDURL(),ident.publickey)


    def _add_fail(x, idurl):
        dhnio.Dprint(10, 'transport_ssh.add_key WARNING adding ssh key for ' + idurl)


    for idurl in idlist:
        if idurl.strip() == '':
            continue

        ident = identitycache.FromCache(idurl)
        if ident == '' or ident is None:
            d = identitycache.immediatelyCaching(idurl)
            d.addCallback(_add_ok, idurl)
            d.addErrback(_add_fail, idurl)
            continue

        add_key(ident.getIDURL(),ident.publickey)




#-------------------------------------------------------------------------------
# SEND
#-------------------------------------------------------------------------------

class SendingFactory(direct.SSHClientFactory):
    def clientConnectionLost(self, connector, reason):
        dhnio.Dprint(14,'transport_ssh.SendingFactory.clientConnectionLost with %s : %s' % (str(connector.getDestination()) ,str(reason.getErrorMessage())))
        direct.SSHClientFactory.clientConnectionLost(self, connector, reason)

    def clientConnectionFailed(self, connector, reason):
##        dhnio.Dprint(1,'transport_ssh.SendingFactory.clientConnectionFailed NETERROR with %s : %s' % (str(connector.getDestination()) ,str(reason.getErrorMessage())))
        direct.SSHClientFactory.clientConnectionFailed(self, connector, reason)
##        transport_control.connectionStatusCallback('ssh', connector.getDestination(), reason, 'connection failed' )
        transport_control.sendStatusReport(connector.getDestination(),
            self.userAuthObject.instance.filename,
            'failed', 'ssh', reason, 'connection failed' )

def send(host, port, username, public_key, private_key, filename):
    def connection_done(x, filename, c):
        dhnio.Dprint(14, 'transport_ssh.send.connection_done ')
##        c.transport.loseConnection()
    def connection_fail(x, filename, c):
        dhnio.Dprint(14, 'transport_ssh.send.connection_fail ')
##        if misc.transport_control_using():
        transport_control.sendStatusReport(host+':'+port, filename,
            'failed', 'ssh',
            x, 'connection failed')      # not so good
##        c.transport.loseConnection()
    def done(x, filename, c):
        dhnio.Dprint(14, 'transport_ssh.send.done')
##        if misc.transport_control_using():
        transport_control.sendStatusReport(host+':'+port, filename,
            'finished', 'ssh')
        c.transport.loseConnection()
    transport_control.log('ssh', 'start sending to '+str(host)+':'+str(port))
    dhnio.Dprint(12, 'transport_ssh.send: %s -> %s:%s' % (filename,host,port))
    port_ = port
    try:
        port_ = int(port)
    except:
        dhnio.Dprint(1,'transport_ssh.send ERROR wrong port number: ' + str(port))
        return
    opts = options.ConchOptions()
    opts.conns = ['direct']
    opts['nocache'] = True
    opts['subsystem'] = 'sftp'
    result_defer = Deferred()
    conn = MySSHConnection(filename, result_defer)
    conn.options = opts
    vhk = my_verifyHostKey
    uao = MyUserAuth(username, conn, public_key, private_key)
    d = Deferred()
    factory = SendingFactory(d, opts, vhk, uao)
    c = reactor.connectTCP(host, port_, factory)
##    client = c.transport
    d.addCallback(connection_done, filename, c)
    d.addErrback(connection_fail, filename, c)
    result_defer.addCallback(done, filename, c)
##    def file_sent(a):
##        dhnio.Dprint(12, 'transport_ssh.send.file_sent: file sent to server')
##        transport_control.log('ssh', 'sending done')
##        client.loseConnection()
##    stopD.addCallback(file_sent)




#-------------------------------------------------------------------------------
# RECEIVE
#-------------------------------------------------------------------------------

def receive(port):
    transport_control.log('ssh', 'start listening on port %s' % str(port))
    dhnio.Dprint(8,'transport_ssh.receive going to listen on port %s' % port)

    def _try_receiving(port, count):
        dhnio.Dprint(10, "transport_ssh.receive count=%d" % count)
        components.registerAdapter(None, MyAvatar, session.ISession)
        d = Deferred()
        myrealm = MyRealm(d)
        portal_ = portal.Portal(myrealm)
        portal_.registerChecker(MyPublicKeyChecker())
        MyFactory.portal = portal_
        try:
            mylistener = reactor.listenTCP(port, MyFactory())
        except:
            mylistener = None
        return mylistener

    def _loop(port, result, count):
        l = _try_receiving(port, count)
        if l is not None:
            dhnio.Dprint(8, "transport_ssh.receive started on port "+ str(port))
            result.callback(l)
            return
        if count > 10:
            dhnio.Dprint(1, "transport_ssh.receive WARNING port %s is busy!" % str(port))
            result.errback(None)
            return
        reactor.callLater(10, _loop, port, result, count+1)

    res = Deferred()
    _loop(port, res, 0)
    return res

    ##        transport_control.connectionStatusCallback('ssh', None, None, 'error listening port %s' % str(port))
##            transport_control.receiveStatusReport('', 'failed',
##                proto='ssh',
##                message='error listening port %s' % str(port))
    ##        dhnio.Dprint(1,'transport_ssh.receive NETERROR listening port %s' % str(port))
##            dhnio.DprintException()

##    def _receive_done(a):
##        print 'receive done. close port.'
##        mylistener.stopListening()
        #we should close port here.
        #but in transport_control we already close it and
        #this line raise error
        #i do not understand this Exception. i am keep thinking
##    return mylistener



def update_db():
    idlist = set()
    idlist.add(misc.getLocalID())
    idlist.add(settings.CentralID())
    #Not sure we need this:
    #idlist.add(settings.MoneyServerID())
    idlist = idlist.union(contacts.getContactsAndCorrespondents())

    add_identity2db(idlist)

def init():
    dhnio.Dprint(4, 'transport_ssh.init')
    update_db()

def main():
    if len(sys.argv) == 4:
        key_pub = dhncrypto.MyPublicKey()
        key_priv = dhncrypto.MyPrivateKey()
        send(sys.argv[1], sys.argv[2], 'testssh', key_pub, key_priv, sys.argv[3])
    elif len(sys.argv) == 3:
        add_key('guestkey', open(sys.argv[2]).read())
        receive(int(sys.argv[1]))
    else:
        print 'args:\n    [host] [port] [file name]\n  or\n    [port] [public key file name]\n'
        sys.exit()

if __name__ == '__main__':
    dhnio.SetDebug(18)
    lf = file('log.p2p', 'ab+')
    log.startLogging(lf, setStdout=False)
    main()
    reactor.run()






#!/usr/bin/python
#restore.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#
#  At least for now we will work one block at a time, though packets in parallel.
#  We ask transport_control for all the data packets for a block then see if we
#  get them all or need to ask for some parity packets.  We do this till we have
#  gotten a block with the "LastBlock" flag set.  If we have tried several times
#  and not gotten data packets from a supplier we can flag him as suspect-bad
#  and start requesting a parity packet to cover him right away.
#
#  When we are missing a data packet we pick a parity packet where we have all the
#  other data packets for that parity so we can recover the missing data packet.
#  This network cost for this is just as low as if we read the data packet.
#  But most of the time we won't bother reading the parities.  Just uses up bandwidth.
#
#    We don't want to fire someone till
#    after we have finished a restore in case we have other problems and they might come
#    back to life and save us.  However, we might keep packets for a node we plan to fire.
#    This would make replacing them very easy.
#
#  At the "tar" level a user will have choice of full restore (to new empty system probably)
#  or to restore to another location on disk, or to just recover certain files.  However,
#  in this module we have to read block by block all of the blocks.
#
#
# How do we restore if we lost everything?
#     Our ( public/private-key and eccmap) could be:
#                  1)  at DHN  (encrypted with pass phrase)
#                  2)  on USB in safe or deposit box   (encrypted with passphrase or clear text)
#                  3)  with friends or scrubbers (encrypted with passphrase)
# The other thing we need is the backupIDs which we can get from our suppliers with the ListFiles command.
# The ID is something like   F200801161206  or I200801170401 indicating full or incremental.

import os
import sys
import time
import gc

try:
    from twisted.internet import reactor
except:
    sys.exit('Error initializing twisted.internet.reactor in restore.py')


from twisted.internet import threads
from twisted.internet.defer import Deferred


import lib.misc as misc
import lib.dhnio as dhnio
import lib.eccmap as eccmap
import lib.dhncrypto as dhncrypto
import lib.settings as settings
import lib.packetid as packetid
import lib.contacts as contacts
import lib.tmpfile as tmpfile
# from lib.automat import Automat
import lib.automat as automat
import lib.automats as automats


import raidread
import dhnblock
import io_throttle
import events
import contact_status


#------------------------------------------------------------------------------ 


class restore(automat.Automat):
    timers = {'timer-01sec':    (0.1, ['RUN',  ]),
              'timer-1sec':     (1,   ['REQUEST',]),
              'timer-5sec':     (5,   ['REQUEST',]), } 

    def __init__(self, BackupID, OutputFile): # OutputFileName 
        self.CreatorID = misc.getLocalID()
        self.BackupID = BackupID
        self.PathID, self.Version = packetid.SplitBackupID(self.BackupID)
        self.File = OutputFile
        # is current active block - so when add 1 we get to first, which is 0
        self.BlockNumber = -1              
        self.OnHandData = []
        self.OnHandParity = []
        self.AbortState = False
        self.Done = False
        self.EccMap = eccmap.Current()
        self.LastAction = time.time()
        self.InboxPacketsQueue = []
        self.InboxQueueWorker = None
        self.InboxQueueDelay = 1
        # For anyone who wants to know when we finish
        self.MyDeferred = Deferred()       
        self.packetInCallback = None
        self.blockRestoredCallback = None
        
        automat.Automat.__init__(self, 'restore', 'AT_STARTUP', 4)
        self.automat('init')
        events.info('restore', '%s start restoring' % self.BackupID)
        # dhnio.Dprint(6, "restore.__init__ %s, ecc=%s" % (self.BackupID, str(self.EccMap)))

    #--- restore() ---
    
    def A(self, event, arg):
        #---AT_STARTUP---
        if self.state is 'AT_STARTUP':
            if event == 'init' :
                self.state = 'RUN'
        #---RUN---
        elif self.state is 'RUN':
            if event == 'timer-01sec' and not self.isAborted(arg) :
                self.state = 'REQUEST'
                self.doStartNewBlock(arg)
                self.doReadPacketsQueue(arg)
                self.doScanExistingPackets(arg)
                self.doRequestPackets(arg)
            elif event == 'timer-01sec' and self.isAborted(arg) :
                self.state = 'ABORTED'
                self.doDeleteAllRequests(arg)
                self.doCloseFile(arg)
                self.doReportAborted(arg)
                self.doDestroyMe(arg)
        #---REQUEST---
        elif self.state is 'REQUEST':
            if event == 'packet-came-in' and self.isPacketValid(arg) and self.isCurrentBlock(arg) :
                self.doSavePacket(arg)
            elif event == 'timer-1sec' and self.isAborted(arg) :
                self.state = 'ABORTED'
                self.doPausePacketsQueue(arg)
                self.doDeleteAllRequests(arg)
                self.doCloseFile(arg)
                self.doReportAborted(arg)
                self.doDestroyMe(arg)
            elif ( event == 'timer-1sec' or event == 'request-done' ) and not self.isAborted(arg) and self.isBlockFixable(arg) :
                self.state = 'RAID'
                self.doPausePacketsQueue(arg)
                self.doReadRaid(arg)
            elif event == 'timer-5sec' and not self.isAborted(arg) and self.isTimePassed(arg) :
                self.doScanExistingPackets(arg)
                self.doRequestPackets(arg)
        #---RAID---
        elif self.state is 'RAID':
            if event == 'raid-done' :
                self.state = 'BLOCK'
                self.doRestoreBlock(arg)
        #---BLOCK---
        elif self.state is 'BLOCK':
            if event == 'block-restored' and self.isBlockValid(arg) and not self.isLastBlock(arg) :
                self.state = 'RUN'
                self.doWriteRestoredData(arg)
                self.doDeleteBlockRequests(arg)
                self.doRemoveTempFile(arg)
            elif event == 'block-restored' and self.isBlockValid(arg) and self.isLastBlock(arg) :
                self.state = 'DONE'
                self.doWriteRestoredData(arg)
                self.doDeleteAllRequests(arg)
                self.doRemoveTempFile(arg)
                self.doCloseFile(arg)
                self.doReportDone(arg)
                self.doDestroyMe(arg)
            elif event == 'block-restored' and not self.isBlockValid(arg) :
                self.state = 'FAILED'
                self.doDeleteAllRequests(arg)
                self.doRemoveTempFile(arg)
                self.doCloseFile(arg)
                self.doReportFailed(arg)
                self.doDestroyMe(arg)
        #---ABORTED---
        elif self.state is 'ABORTED':
            pass
        #---FAILED---
        elif self.state is 'FAILED':
            pass
        #---DONE---
        elif self.state is 'DONE':
            pass

    def isAborted(self, arg):
        return self.AbortState

    def isTimePassed(self, arg):
        return time.time() - self.LastAction > 60
    
    def isPacketValid(self, NewPacket):
        if not NewPacket.Valid():
            return False
        if NewPacket.DataOrParity() not in ['Data', 'Parity']:
            return False
        return True
    
    def isCurrentBlock(self, NewPacket):
        return NewPacket.BlockNumber() == self.BlockNumber
    
    def isBlockFixable(self, arg):
        return self.EccMap.Fixable(self.OnHandData, self.OnHandParity)
    
    def isBlockValid(self, arg):
        NewBlock = arg[0]
        if NewBlock is None:
            return False
        return NewBlock.Valid()
    
    def isLastBlock(self, arg):
        NewBlock = arg[0]
        return NewBlock.LastBlock == str(True)
    
    def doStartNewBlock(self, arg):
        self.LastAction = time.time()
        self.BlockNumber += 1    
        dhnio.Dprint(6, "restore.doStartNewBlock " + str(self.BlockNumber))
        self.OnHandData = [False] * self.EccMap.datasegments
        self.OnHandParity = [False] * self.EccMap.paritysegments

    def doScanExistingPackets(self, arg):
        for SupplierNumber in range(self.EccMap.datasegments):
            PacketID = packetid.MakePacketID(self.BackupID, self.BlockNumber, SupplierNumber, 'Data')
            self.OnHandData[SupplierNumber] = os.path.exists(os.path.join(settings.getLocalBackupsDir(), PacketID))
        for SupplierNumber in range(self.EccMap.paritysegments):
            PacketID = packetid.MakePacketID(self.BackupID, self.BlockNumber, SupplierNumber, 'Parity')
            self.OnHandParity[SupplierNumber] = os.path.exists(os.path.join(settings.getLocalBackupsDir(), PacketID))
        
    def doRequestPackets(self, arg):
        packetsToRequest = []
        for SupplierNumber in range(self.EccMap.datasegments):
            SupplierID = contacts.getSupplierID(SupplierNumber) 
            if not SupplierID:
                continue
            if not self.OnHandData[SupplierNumber] and contact_status.isOnline(SupplierID):
                packetsToRequest.append((SupplierID, packetid.MakePacketID(self.BackupID, self.BlockNumber, SupplierNumber, 'Data')))
        for SupplierNumber in range(self.EccMap.paritysegments):
            SupplierID = contacts.getSupplierID(SupplierNumber) 
            if not SupplierID:
                continue
            if not self.OnHandParity[SupplierNumber] and contact_status.isOnline(SupplierID):
                packetsToRequest.append((SupplierID, packetid.MakePacketID(self.BackupID, self.BlockNumber, SupplierNumber, 'Parity')))
        for SupplierID, packetID in packetsToRequest:
            io_throttle.QueueRequestFile(
                self.PacketCameIn, 
                self.CreatorID, 
                packetID, 
                self.CreatorID, 
                SupplierID)
        dhnio.Dprint(6, "restore.doRequestPackets requested %d packets for block %d" % (len(packetsToRequest), self.BlockNumber))
        del packetsToRequest
        self.automat('request-done')
    
    def doReadRaid(self, arg):
        fd, filename = tmpfile.make('restore', 
            prefix=self.BackupID.replace('/','_')+'_'+str(self.BlockNumber)+'_')
        os.close(fd)
        threads.deferToThread(
            raidread.raidread,
                filename, 
                eccmap.CurrentName(), 
                self.Version, 
                self.BlockNumber, 
                os.path.join(settings.getLocalBackupsDir(), self.PathID) ).addBoth(
                    lambda restored_blocks: self.automat('raid-done', filename))
         
    def doReadPacketsQueue(self, arg):
        reactor.callLater(0, self.ProcessInboxQueue)
    
    def doPausePacketsQueue(self, arg):
        if self.InboxQueueWorker is not None:
            if self.InboxQueueWorker.active():
                self.InboxQueueWorker.cancel()
            self.InboxQueueWorker = None
    
    def doSavePacket(self, NewPacket):
        packetID = NewPacket.PacketID
        pathID, version, packetBlockNum, SupplierNumber, dataORparity = packetid.SplitFull(packetID)
        if dataORparity == 'Data':
            self.OnHandData[SupplierNumber] = True
        elif NewPacket.DataOrParity() == 'Parity':
            self.OnHandParity[SupplierNumber] = True
        filename = os.path.join(settings.getLocalBackupsDir(), packetID)
        dirpath = os.path.dirname(filename)
        if not os.path.exists(dirpath):
            try:
                dhnio._dirs_make(dirpath)
            except:
                dhnio.DprintException()
        # either way the payload of packet is saved
        if not dhnio.WriteFile(filename, NewPacket.Payload):
            dhnio.Dprint(6, "restore.doSavePacket WARNING unable to write to %s" % filename)
            return
        dhnio.Dprint(6, "restore.doSavePacket %s saved" % packetID)
        if self.packetInCallback is not None:
            self.packetInCallback(self.BackupID, NewPacket)
    
    def doRestoreBlock(self, filename):
        blockbits = dhnio.ReadBinaryFile(filename)
        if not blockbits:
            self.automat('block-restored', (None, filename))
            return 
        splitindex = blockbits.index(":")
        lengthstring = blockbits[0:splitindex]
        try:
            datalength = int(lengthstring)                                  # real length before raidmake/ECC
            blockdata = blockbits[splitindex+1:splitindex+1+datalength]     # remove padding from raidmake/ECC
            newblock = dhnblock.Unserialize(blockdata)                      # convert to object
        except:
            datalength = 0
            blockdata = ''
            newblock = None
        self.automat('block-restored', (newblock, filename))
    
    def doWriteRestoredData(self, arg):
        NewBlock = arg[0]
        SessionKey = dhncrypto.DecryptLocalPK(NewBlock.EncryptedSessionKey)
        paddeddata = dhncrypto.DecryptWithSessionKey(SessionKey, NewBlock.EncryptedData)
        newlen = int(NewBlock.Length)
        data = paddeddata[:newlen]
        # Add to the file where all the data is going
        try:
            # self.File.write(data)
            os.write(self.File, data)
        except:
            dhnio.DprintException()
        if self.blockRestoredCallback is not None:
            self.blockRestoredCallback(self.BackupID, NewBlock)
    
    def doDeleteAllRequests(self, arg):
        io_throttle.DeleteBackupRequests(self.BackupID)
                                         
    def doDeleteBlockRequests(self, arg):
        io_throttle.DeleteBackupRequests(self.BackupID + "-" + str(self.BlockNumber))

    def doRemoveTempFile(self, arg):
        tmpfile.throw_out(arg[1], 'block restored')

    def doCloseFile(self, arg):
        # self.File.close()
        os.close(self.File)
    
    def doReportAborted(self, arg):
        dhnio.Dprint(6, "restore.doReportAborted " + self.BackupID)
        self.Done = True
        self.MyDeferred.callback(self.BackupID+' aborted')
        events.info('restore', '%s restoring were aborted' % self.BackupID)
    
    def doReportFailed(self, arg):
        dhnio.Dprint(6, "restore.doReportFailed ERROR - the block does not look good")
        self.Done = True
        self.MyDeferred.errback(Exception(self.BackupID+' failed'))
        events.notify('restore', '%s failed to restore block number %d' % (self.BackupID, self.BlockNumber))
    
    def doReportDone(self, arg):
        # dhnio.Dprint(6, "restore.doReportDone - restore has finished. All is well that ends well !!!")
        self.Done = True
        self.MyDeferred.callback(self.BackupID+' done')
        events.info('restore', '%s restored successfully' % self.BackupID)
    
    def doDestroyMe(self, arg):
        # automats.get_automats_by_index().pop(self.index)
        automat.objects().pop(self.index)
        collected = gc.collect()
        # dhnio.Dprint(6, 'restore.doDestroyMe collected %d objects' % collected)

    #--- restore() ---

    def PacketCameIn(self, NewPacket, state):
        if state == 'received':
            self.InboxPacketsQueue.append(NewPacket)

    def ProcessInboxQueue(self):
        if len(self.InboxPacketsQueue) > 0:
            NewPacket = self.InboxPacketsQueue.pop(0)
            self.automat('packet-came-in', NewPacket)
        self.InboxQueueWorker = reactor.callLater(self.InboxQueueDelay, self.ProcessInboxQueue)
    
    def SetPacketInCallback(self, cb):
        self.packetInCallback = cb

    def SetBlockRestoredCallback(self, cb):
        self.blockRestoredCallback = cb

    def Abort(self): # for when user clicks the Abort restore button on the gui
        dhnio.Dprint(4, "restore.Abort " + self.BackupID)
        self.AbortState = True




from time import time
from nitro.async import AsyncOp
from cspace.network.location import lookupUser

# cache entry states
(ES_DEFAULT,ES_LOOKINGUP,ES_NOTIFYING) = range(3)

class _Entry( object ) :
    def __init__( self, publicKey ) :
        self.publicKey = publicKey
        self.location = None
        self.timestamp = time()
        self.state = ES_DEFAULT
        self.lookupOp = None
        self.notifyOps = None

class LocationCache( object ) :
    def __init__( self, dhtClient, nodeTable, reactor ) :
        self.dhtClient = dhtClient
        self.nodeTable = nodeTable
        self.reactor = reactor
        self.d = {}

    def close( self ) :
        for entry in self.d.values() :
            assert entry.state != ES_NOTIFYING
            if entry.state == ES_LOOKINGUP :
                entry.lookupOp.cancel()
                entry.notifyOps = None
        self.d.clear()

    def _getEntry( self, publicKey ) :
        return self.d.get( publicKey.toDER_PublicKey() )

    def getLocation( self, publicKey ) :
        entry = self._getEntry( publicKey )
        if entry is None :
            return None
        return entry.location

    def _onLookupUser( self, entry, location ) :
        entry.location = location
        entry.timestamp = time()
        entry.lookupOp = None
        entry.state = ES_NOTIFYING
        for op in list(entry.notifyOps) :
            if op in entry.notifyOps :
                entry.notifyOps.remove(op)
                op.notify( location )
        entry.notifyOps = None
        entry.state = ES_DEFAULT

    def refreshUser( self, publicKey, callback=None ) :
        pubKeyData = publicKey.toDER_PublicKey()
        entry = self.d.get( pubKeyData )
        if entry is None :
            entry = _Entry( publicKey )
            self.d[pubKeyData] = entry
        assert entry.state != ES_NOTIFYING
        if entry.state == ES_DEFAULT :
            assert entry.lookupOp is None
            assert entry.notifyOps is None
            def onLookupUser( location ) :
                self._onLookupUser( entry, location )
            entry.lookupOp = lookupUser( publicKey, self.dhtClient,
                    self.nodeTable,
                    lambda loc : self._onLookupUser(entry,loc) )
            entry.state = ES_LOOKINGUP
            entry.notifyOps = set()
        def doCancel() :
            entry.notifyOps.remove( op )
        op = AsyncOp( callback, doCancel )
        entry.notifyOps.add( op )
        return op



#!/usr/bin/python
#
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
# This uses the Linux "dump" and "restore" commands.
#
# These need to be run as root since they access the filesystem via the raw device.
# Tried setting the sticky bit, but even then dump will not open the raw device.
# Can test as user with:
#       /sbin/dump -0 -f - /usr/share/man > /tmp/dump.1
#
# At the moment root does not seem to have the right path for twisted.

import datetime
import popen2
import os

#
#  We want a pipe output or input so we don't need to store intermediate data.
#
#  The popen starts another process.  That process can block but we don't.
#  blockmaker.py only takes data from this pipe when it is ready.

# Returns file descriptor for process that makes tar archive
def backup(directorypath):
    fdout,fdin = popen2.popen2("/sbin/dump -0 -f - %s" % (directorypath))

    print "dump pipe is open"
    return(fdout)



#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.


# __all__ = ["shtoom", "cspace", "cspaceapps", "nitro", ]




#!/usr/bin/env python
#shtoom/stun.py

# This code comes from Shtoom: http://divmod.org/projects/shtoom
# Licensed under the GNU LGPL.
# Copyright (C) 2004 Anthony Baxter
# $Id: stun.py,v 1.15 2004/03/02 14:22:31 anthony Exp $

import sys
import struct, socket, time
try:
    from twisted.internet import reactor
except:
    sys.exit('Error initializing twisted.internet.reactor in shtoom/stun.py')

from twisted.internet import defer
from twisted.internet.protocol import DatagramProtocol
from twisted.python import log
import warnings

import sets


# from shtoom.interfaces import NATMapper as INATMapper

from defcache import DeferredCache
# from shtoom.nat import BaseMapper

STUNVERBOSE = False
# If we're going to follow RFC recommendation, make this 7
MAX_RETRANSMIT = 7

INITIAL_TIMEOUT = 0.100
BACKOFF_TIME = 0.050
MAX_BACKOFF = 4

# Work to be done:
#  - Reverse engineer the C code stun client - why is it giving up so fast?
#  - Do the shared secret stuff? looks like it needs TLS and crap like that.
#  - Move the StunPolicy code to become nat.NATPolicy - we still don't want to
#    use a NAT if both addresses are RFC1918 addresses.
#  - Make the STUN discovery code use a deferred for it's result!
#  - Cache the first responding STUN server after discovery (in the NatType
#    object, maybe?) and only use it? Then need to handle it failing later.
#
#  What happens to any delayed STUN packets that arrive after the Hook gives
#  the transport back? The code that uses the transport should be forgiving,
#  I guess.

# This should be replaced with lookups of
# _stun._udp.divmod.com and _stun._udp.wirlab.net
DefaultServers = [
('stun.ekiga.net', 3478),
##('stun.fwdnet.net', 3478),
('stun.ideasip.com', 3478),
#('stun01.sipphone.com', 3478),
('stun.softjoys.com', 3478),
('stun.voipbuster.com', 3478),
('stun.voxgratia.org', 3478),
('stun.xten.com', 3478),
('stunserver.org', 3478),
('stun.sipgate.net', 10000),

#    ('stun.ekiga.net', 3478),    # added from http://www.voip-info.org/wiki-STUN
#    ('stun.fwdnet.net', 3478),
#    ('stun.ideasip.com', 3478),
#    ('stun.xten.net', 3478),     # old and broken ones
#    ('sip.iptel.org', 3478),
#    ('stun2.wirlab.net', 3478),
#    ('stun.fwdnet.net', 3478),
#    ('stun2.fwdnet.net', 3478),
#    ('stun.wirlab.net', 3478), #
#    ('stun1.vovida.org', 3478), #
#    ('tesla.divmod.net', 3478), #
#    ('erlang.divmod.net', 3478), #
#    ('69.90.168.13', 3478), # stun.fednet.net
#    ('69.90.168.14', 3478), # stun2.fednet.net
#    ('64.69.76.23', 3478), # stun.xten.net
#    ('195.37.77.99', 3478), # sip.iptel.org
#    ('192.98.81.87', 3478), # stun2.wirlab.net
#    ('192.98.81.66', 3478), # stun.wirlab.net
#    ('128.107.250.38', 3478), # stun1.vovida.org
#    ('204.91.10.94', 3478), # tesla.divmod.net
#    ('204.91.10.93', 3478), # erlang.divmod.net
]

StunTypes = {
   0x0001: 'MAPPED-ADDRESS',
   0x0002: 'RESPONSE-ADDRESS ',
   0x0003: 'CHANGE-REQUEST',
   0x0004: 'SOURCE-ADDRESS',
   0x0005: 'CHANGED-ADDRESS',
   0x0006: 'USERNAME',
   0x0007: 'PASSWORD',
   0x0008: 'MESSAGE-INTEGRITY',
   0x0009: 'ERROR-CODE',
   0x000a: 'UNKNOWN-ATTRIBUTES',
   0x000b: 'REFLECTED-FROM',
}

CHANGE_NONE = struct.pack('!i',0)
CHANGE_PORT = struct.pack('!i',2)
CHANGE_IP = struct.pack('!i',4)
CHANGE_BOTH = struct.pack('!i',6)

for k,v in StunTypes.items():
    StunTypes[v] = k
del k, v

class _NatType:
    def __init__(self, name, useful=True, blocked=False):
        self.name = name
        self.useful = useful
        self.blocked = blocked
    def __repr__(self):
        return '<NatType %s>'%(self.name)

NatTypeUDPBlocked = _NatType('UDPBlocked', useful=False, blocked=True)
NatTypeNone = _NatType('None')
NatTypeSymUDP = _NatType('SymUDP')
NatTypeFullCone = _NatType('FullCone')
NatTypeSymmetric = _NatType('Symmetric', useful=False)
NatTypeRestrictedCone = _NatType('RestrictedCone')
NatTypePortRestricted = _NatType('PortRestricted')


# For testing - always return this STUN type
_ForceStunType = None

def hexify(s):
    if s is None:
        return
    ret = ''.join([ '%x'%(ord(c)) for c in s ])
    return ret


import os
if hasattr(os, 'urandom'):
    def getRandomTID():
        return os.urandom(16)
elif os.path.exists('/dev/urandom'):
    def getRandomTID():
        return open('/dev/urandom').read(16)
else:
    def getRandomTID():
        # It's not absolutely necessary to have a particularly strong TID here
        import random
        tid = [ chr(random.randint(0,255)) for x in range(16) ]
        tid = ''.join(tid)
        return tid

def _parseStunResponse(dgram, address, expectedTID=None, oldtids=[]):
    mt, pktlen, tid = struct.unpack('!hh16s', dgram[:20])
    if expectedTID is not None and expectedTID != tid:
        # a response from an earlier request
        if tid in oldtids:
            # discard
            if STUNVERBOSE:
                log.msg("ignoring belated STUN response to %r from %s"%(
                            hexify(tid), repr(address)), system='stun')
            return
        log.msg("got unexpected STUN response %r != %r from %s"%
                        (hexify(expectedTID), hexify(tid),
                        repr(address),), system='stun')
        return
    resdict = {}
    if mt == 0x0101:
        log.msg("got STUN response to %r from %s"%(hexify(expectedTID),
                                                   repr(address)),
                                                        system='stun')
        # response
        remainder = dgram[20:]
        while remainder:
            avtype, avlen = struct.unpack('!hh', remainder[:4])
            val = remainder[4:4+avlen]
            avtype = StunTypes.get(avtype, '(Unknown type %04x)'%avtype)
            remainder = remainder[4+avlen:]
            if avtype in ('MAPPED-ADDRESS',
                          'CHANGED-ADDRESS',
                          'SOURCE-ADDRESS'):
                dummy,family,port,addr = struct.unpack('!ccH4s', val)
                addr = socket.inet_ntoa(addr)
                # if STUNVERBOSE:
                #     print avtype, addr, port
                if avtype == 'MAPPED-ADDRESS':
                    resdict['externalAddress'] = (addr, port)
                elif avtype == 'CHANGED-ADDRESS':
                    resdict['_altStunAddress'] = (addr, address[1])
#                elif avtype == 'SOURCE-ADDRESS':
#                    resdict['sourceAddress'] = (addr, port)
                elif address[0] != addr:
                    # Some son of a bitch is rewriting packets on the way
                    # back. AAARGH.
                    log.msg('WARNING: packets are being rewritten %r != %r'%
                            (address, (addr,port)), system='stun')
                    return
            else:
                log.msg("STUN: unhandled AV %s, val %r"%(avtype,
                                                         repr(val)),
                                                         system='stun')
    elif mt == 0x0111:
        log.err("STUN got an error response")
    return resdict

class  _StunBase(object):

    def sendRequest(self, server, tid=None, avpairs=()):
        # <AP>
        if not self.transport:
##            print "No transport defined, cannot send STUN request"
            return
        # </AP>
        if tid is None:
            tid = getRandomTID()
        mt = 0x1 # binding request
        avstr = ''
        # add any attributes
        if not avpairs:
            avpairs = ('CHANGE-REQUEST', CHANGE_NONE),
        for a,v in avpairs:
            avstr = avstr + struct.pack('!hh', StunTypes[a], len(v)) + v
        pktlen = len(avstr)
        if pktlen > 65535:
            raise ValueError, "stun request too big (%d bytes)"%pktlen
        pkt = struct.pack('!hh16s', mt, pktlen, tid) + avstr
        if STUNVERBOSE:
            print "sending request %r with %d avpairs to %r (in state %s)"%(
                            hexify(tid), len(avpairs), server, self._stunState)
        try:
            self.transport.write(pkt, server)
        except:
            if STUNVERBOSE:
                print "exception during sending request"

class StunDiscoveryProtocol(DatagramProtocol, _StunBase):

    stunDiscoveryRetries = 0

    def __init__(self, servers=DefaultServers, *args, **kwargs):
       # Potential STUN servers
        self._potentialStuns = {}
        # See flowchart ascii art at bottom of file.
        self._stunState = '1'
        self._finished = False
        self._altStunAddress = None
        self.externalAddress = None
        self.localAddress = None
        self.expectedTID = None
        self.oldTIDs = sets.Set()
        self.natType = None
        self.timerTask = None
        self.timeout = 20
        self.servers = [(host, port) for host, port in servers]
        # super(StunDiscoveryProtocol, self).__init__(*args, **kwargs)

    def stateChanged(self, old, new):
        if STUNVERBOSE:
            print "stun state changed: [%s] -> [%s]" % (old, new)

    def initialStunRequest(self, address):
##        print 'initialStunRequest', address
        # <AP>
        if self._finished:
            return
        # </AP>
        tid = getRandomTID()
        delayed = reactor.callLater(INITIAL_TIMEOUT,
                                    self.retransmitInitial, address, tid)
        self._potentialStuns[tid] = delayed
        self.oldTIDs.add(tid)
        self.sendRequest(address, tid=tid)

    def retransmitInitial(self, address, tid, count=1):
        # <AP>
        if self._finished:
            return
        # </AP>
        if count <= MAX_RETRANSMIT:
            t = BACKOFF_TIME * 2**min(count, MAX_BACKOFF)
            delayed = reactor.callLater(t, self.retransmitInitial,
                                                address, tid, count+1)
            self._potentialStuns[tid] = delayed
            self.sendRequest(address, tid=tid)
        else:
            if STUNVERBOSE:
                print "giving up on %r"%(address,)
            del self._potentialStuns[tid]
            if not self._potentialStuns:
                if STUNVERBOSE:
                    print "stun state 1 timeout - no internet UDP possible"
                self.natType = NatTypeUDPBlocked
                self._finishedStun()

    def datagramReceived(self, dgram, address):
        if self._finished:
            return
        mt, pktlen, tid = struct.unpack('!hh16s', dgram[:20])
        # Check tid is one we sent and haven't had a reply to yet
        if tid in self._potentialStuns:
            delayed = self._potentialStuns.get(tid)
            if delayed is not None:
                delayed.cancel()
            del self._potentialStuns[tid]
            if self._stunState == '1':
                # We got a (potentially) working STUN server!
                # Cancel the retransmit timers for the other ones
                for k in self._potentialStuns.keys():
                    if self._potentialStuns.has_key(k) and self._potentialStuns[k] is not None:
                        self._potentialStuns[k].cancel()
                        self._potentialStuns[k] = None
                resdict = _parseStunResponse(dgram, address, self.expectedTID,
                                                self.oldTIDs)
                if not resdict:
                    return
                self.handleStunState1(resdict, address)
            else:
                # We already have a working STUN server to play with.
                pass
            return
        resdict = _parseStunResponse(dgram, address, self.expectedTID,
                                                self.oldTIDs)
        if not resdict:
            return
        if STUNVERBOSE:
            print 'calling handleStunState%s'%(self._stunState)
        getattr(self, 'handleStunState%s'%(self._stunState))(resdict, address)

    def handleStunState1(self, resdict, address):
        self.__dict__.update(resdict)

        if self.externalAddress and self._altStunAddress:
            if self.localAddress == self.externalAddress[0]:
                self.stateChanged(self._stunState, '2a')
                self._stunState = '2a'
            else:
                self.stateChanged(self._stunState, '2b')
                self._stunState = '2b'
            self.expectedTID = tid = getRandomTID()
            self.oldTIDs.add(tid)
            self.state2DelayedCall = reactor.callLater(INITIAL_TIMEOUT,
                                                self.retransmitStunState2,
                                                address, tid)
            self.sendRequest(address, tid, avpairs=(
                                    ('CHANGE-REQUEST', CHANGE_BOTH),))

    def handleStunState2a(self, resdict, address):
        self.state2DelayedCall.cancel()
        del self.state2DelayedCall
        if STUNVERBOSE:
            print "2a", resdict
        self.natType = NatTypeNone
        self._finishedStun()

    def handleStunState2b(self, resdict, address):
        self.state2DelayedCall.cancel()
        del self.state2DelayedCall
        if STUNVERBOSE:
            print "2b", resdict
        self.natType = NatTypeFullCone
        self._finishedStun()

    def retransmitStunState2(self, address, tid, count=1):
        # <AP>
        if self._finished:
            return
        # </AP>
        if count <= MAX_RETRANSMIT:
            t = BACKOFF_TIME * 2**min(count, MAX_BACKOFF)
            self.state2DelayedCall = reactor.callLater(t,
                                                    self.retransmitStunState2,
                                                    address, tid, count+1)
            self.sendRequest(address, tid, avpairs=(
                                    ('CHANGE-REQUEST', CHANGE_BOTH),))
        elif self._stunState == '2a':
            self.natType = NatTypeSymUDP
            self._finishedStun()
        else: # 2b
            # Off to state 3 we go!
            self.stateChanged(self._stunState, '3')
            self._stunState = '3'
            self.state3DelayedCall = reactor.callLater(INITIAL_TIMEOUT,
                                                    self.retransmitStunState3,
                                                    address, tid)
            self.expectedTID = tid = getRandomTID()
            self.oldTIDs.add(tid)
            self.sendRequest(self._altStunAddress, tid)

    def handleStunState3(self, resdict, address):
        self.state3DelayedCall.cancel()
        del self.state3DelayedCall
        if STUNVERBOSE:
            print "3", resdict
        if self.externalAddress == resdict['externalAddress']:
            # State 4! wheee!
            self.stateChanged(self._stunState, '4')
            self._stunState = '4'
            self.expectedTID = tid = getRandomTID()
            self.oldTIDs.add(tid)
            self.state4DelayedCall = reactor.callLater(INITIAL_TIMEOUT,
                                                self.retransmitStunState4,
                                                address, tid)
            self.expectedTID = tid = getRandomTID()
            self.oldTIDs.add(tid)
            self.sendRequest(address, tid, avpairs=(
                                    ('CHANGE-REQUEST', CHANGE_PORT),))
        else:
            self.natType = NatTypeSymmetric
            self._finishedStun()

    def retransmitStunState3(self, address, tid, count=1):
        # <AP>
        if self._finished:
            return
        # </AP>
        if count <= (2 * MAX_RETRANSMIT):
            t = BACKOFF_TIME * 2**min(count, MAX_BACKOFF)
            self.state3DelayedCall = reactor.callLater(t,
                                                    self.retransmitStunState3,
                                                    address, tid, count+1)
            self.sendRequest(self._altStunAddress, tid)
        else:
            log.err("STUN Failed in state 3, retrying")
            # We should do _something_ here. a new type BrokenNAT?
            self.stunDiscoveryRetries = self.stunDiscoveryRetries + 1
            if self.stunDiscoveryRetries < 5:
                reactor.callLater(0.2, self.startDiscovery)

    def handleStunState4(self, resdict, address):
        self.state4DelayedCall.cancel()
        del self.state4DelayedCall
        self.natType = NatTypeRestrictedCone
        self._finishedStun()

    def retransmitStunState4(self, address, tid, count = 1):
        # <AP>
        if self._finished:
            return
        # </AP>
        if count < MAX_RETRANSMIT:
            t = BACKOFF_TIME * 2**min(count, MAX_BACKOFF)
            self.state4DelayedCall = reactor.callLater(t,
                                                    self.retransmitStunState4,
                                                    address, tid, count+1)
            self.sendRequest(address, tid, avpairs=(
                                    ('CHANGE-REQUEST', CHANGE_PORT),))
        else:
            self.natType = NatTypePortRestricted
            self._finishedStun()


    def _finishedStun(self):
        self._finished = True
        self.finishedStun()

    def finishedStun(self):
        # Override in a subclass
        if STUNVERBOSE:
            print "firewall type is", self.natType

    def startDiscovery(self):
        from nat import isBogusAddress, getLocalIPAddress
        if _ForceStunType is not None:
            self.natType = _ForceStunType
            reactor.callLater(0, self._finishedStun)
            return
        if not self.transport:
            self.natType = _ForceStunType
            reactor.callLater(0, self._finishedStun)
            return
        localAddress = self.transport.getHost().host
        if STUNVERBOSE:
            print 'startDiscovery, localAddres =', localAddress
        if isBogusAddress(localAddress):
            d = getLocalIPAddress()
            d.addCallback(self._resolveStunServers)
            d.addErrback(self._hostNotResolved)
        else:
            self._resolveStunServers(localAddress)
        self.timerTask = reactor.callLater(self.timeout, self._timeout)

    def _hostNotResolved(self, x):
        if STUNVERBOSE:
            print '_hostNotResolved', x

    def _resolveStunServers(self, localAddress):
        self.localAddress = localAddress
        # reactor.resolve the hosts!
        for host, port in self.servers:
##            print '_resolveStunServers', host, port
            d = reactor.resolve(host)
            d.addCallback(lambda x,p=port: self.initialStunRequest((x, p)))
            d.addErrback(self._hostNotResolved)

    def _timeout(self):
#        if STUNVERBOSE:
#            print '_timeout'
        self._hostNotResolved('timeout %d seconds' % self.timeout)

# class StunHook(_StunBase):
#     """Hook a StunHook into a UDP protocol object, and it will discover
#        STUN settings for it.
#
#        You should probably use the NATMapper approach rather than using
#        StunHook directly.
#     """
#     def __init__(self, prot, servers=DefaultServers, *args, **kwargs):
#         self._protocol = prot
#         self._pending = {}
#         self.servers = servers
#         self.expectedTID = None
#         self.oldTIDs = sets.Set()
#         self._stunState = 'hook'
#         super(StunHook, self).__init__(*args, **kwargs)
#
#     def initialStunRequest(self, address):
#         tid = getRandomTID()
#         self.oldTIDs.add(tid)
#         delayed = reactor.callLater(INITIAL_TIMEOUT,
#                                     self.retransmitInitial, address, tid)
#         self._pending[tid] = delayed
#         self.sendRequest(address, tid=tid)
#
#     def retransmitInitial(self, address, tid, count=1):
#         if count <= MAX_RETRANSMIT:
#             t = BACKOFF_TIME * 2**min(count, MAX_BACKOFF)
#             delayed = reactor.callLater(t, self.retransmitInitial,
#                                             address, tid, count+1)
#             self._pending[tid] = delayed
#             self.sendRequest(address, tid=tid)
#         else:
#             if STUNVERBOSE:
#                 print "giving up on %r"%(address,)
#             del self._potentialStuns[tid]
#             if not self._potentialStuns:
#                 if STUNVERBOSE:
#                     print "stun state 1 timeout - no internet UDP possible"
#                 self.natType = NatTypeUDPBlocked
#                 self.finishedStun()
#
#     def datagramReceived(self, dgram, address):
#         if STUNVERBOSE:
#             print "hook got a datagram from", address
#         if self.deferred is None:
#             # We're already done
#             return
#
#         mt, pktlen, tid = struct.unpack('!hh16s', dgram[:20])
#         if self._pending.has_key(tid):
#             delayed = self._pending[tid]
#             if delayed is not None:
#                 delayed.cancel()
#             del self._pending[tid]
#             resdict = _parseStunResponse(dgram, address,
#                                                 oldtids=self.oldTIDs)
#             if not resdict or not resdict.get('externalAddress'):
#                 # Crap response, ignore it.
#                 return
#
#             # Got a valid response. Clean up around here first.
#             self.uninstallStun()
#             # kill any pending retransmits
#             for delayed in self._pending.values():
#                 if delayed is not None:
#                     delayed.cancel()
#             # send response
#             d, self.deferred = self.deferred, None
#             d.callback(resdict['externalAddress'])
#
#     def installStun(self):
#         self._protocol._mp_datagramReceived = self._protocol.datagramReceived
#         self._protocol.datagramReceived = self.datagramReceived
#         self.transport = self._protocol.transport
#
#     def discoverAddress(self):
#         """ Sniff out external address. Returns a deferred with the external
#             address as a 2-tuple (ip, port)
#         """
#         from twisted.internet import defer
#         self.installStun()
#         self.deferred = defer.Deferred()
#         for host, port in self.servers[:3]:
#             d = reactor.resolve(host)
#             d.addCallback(lambda x, p=port: self.initialStunRequest((x, p)))
#         return self.deferred
#
#     def uninstallStun(self):
#         self._protocol.datagramReceived = self._protocol._mp_datagramReceived
#         del self.transport

class _DetectSTUNProt(StunDiscoveryProtocol):
    d = None
    def finishedStun(self):
        self.d.callback(self.natType)

_cached_stuntype = None

def _getSTUN():
    #print "getSTUN triggering startDiscovery!"
    if _cached_stuntype is not None:
        return defer.succeed(_cached_stuntype)
    stunClient = _DetectSTUNProt()
    stunClient.d = defer.Deferred()
    l = reactor.listenUDP(0, stunClient)
    reactor.callLater(0, stunClient.startDiscovery)
    def _stundone(x, l=l):
        global _cached_stuntype
        _cached_stuntype = x
        l.stopListening()
        return x
    stunClient.d.addCallback(_stundone)
    return stunClient.d

getSTUN = DeferredCache(_getSTUN)

# _cached_mapper = None
# def getMapper():
#     global _cached_mapper
#     if _cached_mapper is None:
#         _cached_mapper = STUNMapper()
#     return _cached_mapper
#
# def clearCache():
#     global _cached_mapper, _cached_stuntype
#     _cached_stuntype = None
#     _cached_mapper = None
#     getSTUN.clearCache()
#
# class STUNMapper(BaseMapper):
#     __implements__ = INATMapper
#     _ptypes = [ 'UDP', ]
#     def __init__(self):
#         self._mapped = {}
#
#     def map(self, port):
#         "See shtoom.interfaces.NATMapper.map"
#         self._checkValidPort(port)
#         cd = defer.Deferred()
#         self._mapped[port] = cd
#         d = getSTUN()
#         d.addCallback(lambda x: self._cb_map_gotSTUN(x, port))
#         return cd
#     map = DeferredCache(map, inProgressOnly=True)
#
#     def _cb_map_gotSTUN(self, stun, port):
#         if not stun.useful:
#             cd = self._mapped[port]
#             del self._mapped[port]
#             cd.errback(ValueError('%r means STUN is useless'%(stun,)))
#             return
#         SH = StunHook(port.protocol)
#         d = SH.discoverAddress()
#         d.addCallback(lambda x: self._cb_map_discoveredAddress(x, port))
#     map = DeferredCache(map, inProgressOnly=True)
#
#     def _cb_map_discoveredAddress(self, addr, port):
#         cd = self._mapped[port]
#         self._mapped[port] = addr
#         cd.callback(addr)
#
#     def info(self, port):
#         "See shtoom.interfaces.NATMapper.info"
#         if port in self._mapped:
#             return self._mapped[port]
#         else:
#             raise ValueError('Port %r is not currently mapped'%(port))
#
#     def unmap(self, port):
#         "See shtoom.interfaces.NATMapper.unmap"
#         # A no-op for STUN
#         if port not in self._mapped:
#             raise ValueError('Port %r is not currently mapped'%(port))
#         del self._mapped[port]
#         return defer.succeed(None)


if __name__ == "__main__":
    STUNVERBOSE=True
    log.FileLogObserver.timeFormat = "%H:%M:%S"
    import sys
    class TestStunDiscoveryProtocol(StunDiscoveryProtocol):

        def finishedStun(self):
            print "STUN finished, results:"
            print "You're behind a %r"%(self.natType)
            if self.natType is NatTypeSymmetric:
                print "You're going to have to use an outbound proxy"
            else:
                print "and external address is %r" % (self.externalAddress,)
            reactor.stop()


    stunClient = TestStunDiscoveryProtocol()
    log.startLogging(sys.stdout)
    try:
        reactor.listenUDP(5061, stunClient)
    except:
        reactor.listenUDP(0, stunClient)
    #reactor.callLater(20, stunClient.stunTimedOut)
    reactor.callLater(0, stunClient.startDiscovery)
    reactor.run()



"""

Copied from RFC 3489


                        +--------+
                        |  Test  | S:1
                        |   I    |
                        +--------+
                             |
                             |
                             V             S:1a
                            /\              /\
                         N /  \ Y          /  \ Y             +--------+
          UDP     <-------/Resp\--------->/ IP \------------->|  Test  | S:2a
          Blocked         \ ?  /          \Same/              |   II   |
                           \  /            \? /               +--------+
                            \/              \/                    |
                                             | N                  |
                                             |                    V
                                             V                    /\
                                         +--------+  Sym.      N /  \
                                    S:2b |  Test  |  UDP    <---/Resp\
                                         |   II   |  Firewall   \ ?  /
                                         +--------+              \  /
                                             |                    \/
                                             V                     |Y
                  /\           S:3           /\                    |
   Symmetric  N  /  \       +--------+   N  /  \                   V
      NAT  <--- / IP \<-----|  Test  |<--- /Resp\               Open
                \Same/      |   III  |     \ ?  /               Internet
                 \? /       +--------+      \  /
                  \/                         \/
                  |                           |Y
                  |                           |
                  |                           V
                  |                           Full
                  |                           Cone
                  V              /\
              +--------+        /  \ Y
         S:4  |  Test  |------>/Resp\---->Restricted
              |    IV  |       \ ?  /
              +--------+        \  /
                                 \/
                                  |N
                                  |       Port
                                  +------>Restricted

                 Figure 2: Flow for type discovery process

Test I: A Binding Request with no change port/change ip
Test II: A Binding Request with change port & change ip
Test III: A Binding Request with no change port/change ip, sent to the
          address from CHANGED-ADDRESS in response to Test I.
Test IV: A Binding Request with change port

"""



import os, sys, logging
from bisect import bisect
try:
    from PyQt4 import QtCore, QtGui
    from PyQt4.QtCore import pyqtSignature, SIGNAL, QSize
    from PyQt4.QtGui import QApplication, qApp, QMessageBox, \
        QIcon, QMainWindow, QDialog, QListWidgetItem, \
        QMenu, QAction, QPixmap
except:
    pass

try :
    from qtrayicon import TrayIcon
    HAS_TRAY = True
except ImportError :
    HAS_TRAY = False

try:
    from nitro.qt4reactor import Qt4Reactor
except:
    pass

from cspace.util.spawn import spawnProcess
from cspace.util.eventer import Eventer
from cspace.util.delaygc import initdelaygc
from cspace.main.common import SETTINGS_VERSION, localSettings
from cspace.main.profile import listProfiles, loadProfile, \
        createProfile, saveProfileContacts
from cspace.main.session import UserSession
from cspace.main.appletserver import AppletServer
from cspace.main.autoupdater import currentBuildNumber, AutoUpdater
try:
    from cspace.main.dialogs import CreateKeyDialog, CreateKeyDoneDialog, \
            GoOnlineDialog, KeyInfoDialog, AddContactDialog, \
            ContactInfoDialog, PermissionsDialog, UpdateNotifyWindow
    import cspace.main.ui.images_rc
    from cspace.main.ui.Ui_MainWindow import Ui_MainWindow
except:
    pass

CONTACT_PROBE_INTERVAL = 4*60
CONTACT_NEXT_PROBE_INTERVAL = 30
execFileAfterExit = None

class Dummy : pass

class ActionManager( object ) :
    def __init__( self, contextMenu ) :
        self.contextMenu = contextMenu
        self.nextActionId = 0
        self.actions = []
        self.defaultActionId = -1

    def registerAction( self, action, actionCallback, actionOrder ) :
        if not self.actions :
            self.contextMenu.insertSeparator( self.contextMenu.actions()[0] )
        actionId = self.nextActionId
        self.nextActionId += 1
        info = [actionOrder,action,actionId,actionCallback,-1]
        pos = bisect( self.actions, info )
        self.actions.insert( pos, info )
        actionObj = QAction( action, None )
        info[4] = actionObj
        self.contextMenu.insertAction( self.contextMenu.actions()[pos], actionObj )
        return actionId

    def setDefaultAction( self, actionId ) :
        if self.defaultActionId >= 0 : return False
        self.defaultActionId = actionId
        return True

    def unregisterAction( self, actionId ) :
        for pos,info in enumerate(self.actions) :
            if info[2] == actionId :
                self.contextMenu.removeAction( info[4] )
                del self.actions[pos]
                if not self.actions :
                    self.contextMenu.removeAction( self.contextMenu.actions()[0] )
                if self.defaultActionId == actionId :
                    self.defaultActionId = -1
                return True
        assert False

    def execAction( self, actionObj, contactName ) :
        for info in self.actions :
            if info[4] is actionObj :
                info[3]( contactName )
                return

    def execDefaultAction( self, contactName ) :
        if self.defaultActionId < 0 : return
        for info in self.actions :
            if info[2] == self.defaultActionId :
                info[3]( contactName )
                return

class Reconnector( object ) :
    def __init__( self, profile, timerCallback, reactor ) :
        self.reconnecting = False
        self.profile = profile
        self.timerCallback = timerCallback 
        self.reactor = reactor
        self.timerOp = None

    def startReconnectTimer( self, errorMsg ) :
        self.errorMsg = errorMsg
        self.reconnecting = True
        self.timeLeft = 20
        self.timerOp = self.reactor.addTimer( 1, self._onTimer )

    def shutdown( self ) :
        if self.timerOp :
            self.timerOp.cancel()
            self.timerOp = None
        self.reconnecting = False

    def _onTimer( self ) :
        self.timeLeft -= 1
        if self.timeLeft == 0 :
            self.shutdown()
        self.timerCallback()

class StatusChecker( object ) :
    def __init__( self, contactsView, session, eventer, reactor ) :
        self.contactsView = contactsView
        self.session = session
        self.reactor = reactor
        self.checkOps = {}
        self.timerOp = None
        def onRemoveContact( item, contact ) :
            self.stop( item )
        eventer.register( 'contact.remove', onRemoveContact )

    def initialize( self ) :
        obj = Dummy()
        def count() : return self.contactsView.count()
        def checking( i ) : return self.contactsView.item(i) in self.checkOps
        def pollTimer() :
            self.timerOp = self.reactor.callLater(
                    CONTACT_NEXT_PROBE_INTERVAL, onTimer )
        def checkNext() :
            if obj.i < count() :
                self.check( self.contactsView.item(obj.i) )
                pollTimer()
                return
            obj.i = -1
            self.timerOp = self.reactor.callLater(
                    CONTACT_PROBE_INTERVAL, onTimer )
        def onTimer() :
            self.timerOp = None
            if obj.i < 0 :
                obj.i = 0
                checkNext()
                return
            if (obj.i < count()) and checking(obj.i) :
                pollTimer()
                return
            obj.i += 1
            checkNext()
        obj.i = -1
        self.timerOp = self.reactor.callLater(
                CONTACT_PROBE_INTERVAL, onTimer )

    def shutdown( self ) :
        if self.timerOp :
            self.timerOp.cancel()
            self.timerOp = None
        for op in self.checkOps.values() :
            op.cancel()
        self.checkOps.clear()

    def check( self, item ) :
        if item in self.checkOps : return
        name = str( item.text() )
        contact = self.session.getProfile().getContactByName( name )
        assert contact
        def onProbe( result ) :
            del self.checkOps[item]
            if result : img = 'user_online.png'
            else : img = 'user_offline.png'
            item.setIcon( QIcon(':/images/%s'%img) )
        op = self.session.probeUserOnline( contact.publicKey, onProbe )
        item.setIcon( QIcon(':/images/user_checking.png') )
        self.checkOps[item] = op 

    def stop( self, item ) :
        op = self.checkOps.pop( item, None )
        if op is not None :
            op.cancel()
        if item in self.checkOps :
            del self.checkOps[item]


class MainWindow( QMainWindow ) :
    OFFLINE = UserSession.OFFLINE
    CONNECTING = UserSession.CONNECTING
    ONLINE = UserSession.ONLINE
    DISCONNECTING = UserSession.DISCONNECTING

    def __init__( self, seedNodes, reactor ) :
        QMainWindow.__init__( self )
        self.ui = Ui_MainWindow()
        self.ui.setupUi( self )
        self._setupIcons()
        self.seedNodes = seedNodes
        self.reactor = reactor
        self.session = UserSession( seedNodes, reactor )
        self.sm = self.session.sm
        self.ev = Eventer()
        self.profile = None
        self.baseTitle = str(self.windowTitle())
        self.exitAfterOffline = False
        self.reconnector = None
        self.keyInfoDialog = None
        self.addContactDialog = None
        self.sm.insertCallback( self._onStateChange )
        self.connect( self.ui.createKeyButton, SIGNAL('clicked()'), self.ui.actionCreateKey.trigger )
        self.connect( self.ui.createKeyButton1, SIGNAL('clicked()'), self.ui.actionCreateKey.trigger )
        self.connect( self.ui.goOnlineButton, SIGNAL('clicked()'), self.ui.actionGoOnline.trigger )
        self.ui.contacts.setWrapping( True )
        self.connect( self.ui.contacts, SIGNAL('customContextMenuRequested(const QPoint&)'),
                self._onContactsContextMenuRequested )
        self.connect( self.ui.contacts, SIGNAL('itemDoubleClicked(QListWidgetItem*)'),
                self._onContactDoubleClicked )
        self.contextMenu = QMenu()
        self.contextMenu.addAction( self.ui.actionCheckStatus )
        self.contextMenu.addAction( self.ui.actionRemoveContact )
        self.contextMenu.addAction( self.ui.actionContactInfo )
        self.connect( self.contextMenu, SIGNAL('triggered(QAction*)'), self._onContextMenuAction )
        self.actionManager = ActionManager( self.contextMenu )
        self.appletServer = AppletServer( self.session, self.actionManager, self.reactor )
        self.appletServer.registerServices()
        self.appletServer.registerActions()
        self.statusChecker = StatusChecker( self.ui.contacts, self.session, self.ev, self.reactor )
        self.contactInfoDialogs = {}
        self.trayMenu = QMenu( self )
        self.trayMenu.addAction( self.ui.actionExit )
        if HAS_TRAY :
            self.trayIcon = TrayIcon( QPixmap(':/images/cspace_offline.png'), 'CSpace', self.trayMenu, self )
            self.connect( self.trayIcon, SIGNAL('clicked(const QPoint&,int)'), self._onTrayClicked )
            self.trayIcon.show()
        else :
            self.trayIcon = None
        self.editPermissionsDialog = None
        self.autoUpdater = AutoUpdater( self.reactor )
        self.autoUpdater.setUpdateCallback( self._onUpdateDownloaded )
        self._updateUI()
        self._doCheckSettings()

    def _setupIcons( self ) :
        def makeIcon( name ) :
            icon = QIcon()
            for res in '16 22 32'.split() :
                icon.addFile( ':/images/%s%s.png' % (name,res) )
            return icon
        actionIcons = (
                'CreateKey register',
                'GoOnline connect',
                'GoOffline disconnect',
                'Exit exit',
                'AddContact user_add',
                'RefreshStatus refresh',
                'CheckStatus refresh',
                'ContactInfo contact_info',
                'RemoveContact user_remove',
                'EditPermissions edit_permissions',
                'KeyInfo key_info' )
        for s in actionIcons :
            name,icon = s.split()
            action = getattr( self.ui, 'action%s'%name )
            action.setIcon( makeIcon(icon) )

    def _updateUI( self ) :
        state = self.sm.current()
        online = (state == self.ONLINE)
        offline = (state == self.OFFLINE)
        connecting = (state == self.CONNECTING)
        disconnecting = (state == self.DISCONNECTING)
        self.ui.actionCreateKey.setEnabled( offline )
        self.ui.actionGoOnline.setEnabled( offline )
        self.ui.actionGoOffline.setEnabled( online )
        self.ui.actionKeyInfo.setEnabled( online )
        self.ui.actionExit.setEnabled( True )
        self.ui.actionAddContact.setEnabled( online )
        self.ui.actionRefreshStatus.setEnabled( online )
        self.ui.actionCheckStatus.setEnabled( online )
        self.ui.actionContactInfo.setEnabled( online )
        self.ui.actionRemoveContact.setEnabled( online )
        self.ui.actionEditPermissions.setEnabled( online )
        self.ui.actionAboutCSpace.setEnabled( True )
        if online : msg = 'Online'
        if offline : msg = 'Offline'
        if connecting : msg = 'Going Online...'
        if disconnecting : msg = 'Going Offline...'
        if not offline :
            msg = '%s - %s' % (self.profile.name,msg)
        self.setWindowTitle( ' - '.join([msg,self.baseTitle]) )
        self.ui.createKeyButton.setEnabled( offline )
        self.ui.createKeyButton1.setEnabled( offline )
        self.ui.goOnlineButton.setEnabled( offline )
        self.ui.connectCancelButton.setEnabled( connecting or disconnecting )
        if online :
            self.ui.stack.setCurrentWidget( self.ui.contactsPage )
        elif offline :
            if self.reconnector :
                assert self.reconnector.reconnecting
                self.ui.connectCancelButton.setEnabled( True )
                self.ui.stack.setCurrentWidget( self.ui.connectingPage )
                self.ui.connectStatus.setText('<b>%s<br/>Reconnecting in %d second(s)...</b>'
                        % (self.reconnector.errorMsg, self.reconnector.timeLeft) )
            elif listProfiles() :
                self.ui.stack.setCurrentWidget( self.ui.offlinePage )
            else :
                self.ui.stack.setCurrentWidget( self.ui.offlineNoUsersPage )
        elif connecting :
            self.ui.stack.setCurrentWidget( self.ui.connectingPage )
            self.ui.connectStatus.setText( '<b>Connecting...</b>' )
        elif disconnecting :
            self.ui.stack.setCurrentWidget( self.ui.connectingPage )
            self.ui.connectStatus.setText( '<b>Disconnecting...</b>' )

    def _onStateChange( self ) :
        if self.sm.previous() == self.ONLINE :
            self.statusChecker.shutdown()
            self.appletServer.clearConnections()
            for dlg in self.contactInfoDialogs.values() :
                dlg.close()
            assert not self.contactInfoDialogs
            if self.editPermissionsDialog :
                self.editPermissionsDialog.close()
            assert not self.editPermissionsDialog
            self.ui.contacts.clear()
            if self.trayIcon is not None :
                self.trayIcon.setIcon( QPixmap(':/images/cspace_offline.png') )
                self.trayIcon.setToolTip( 'CSpace' )
        if self.sm.current() == self.OFFLINE :
            self.profile = None
            if self.reconnector :
                if self.sm.previous() == self.CONNECTING :
                    msg = 'Connect failed.'
                else :
                    msg = 'Disconnected.'
                self.reconnector.startReconnectTimer( msg )
            if self.exitAfterOffline :
                self.reactor.callLater( 0, self.ui.actionExit.trigger )
        elif self.sm.current() == self.ONLINE :
            self.ui.contacts.clear()
            for contactName in self.profile.getContactNames() :
                contact = self.profile.getContactByName( contactName )
                self._addContactItem( contact )
            self.statusChecker.initialize()
            if self.trayIcon is not None :
                self.trayIcon.setIcon( QPixmap(':/images/cspace_online.png') )
                self.trayIcon.setToolTip( 'CSpace - %s' % self.profile.name )
        self._updateUI()

    def _doCheckSettings( self ) :
        st = localSettings()
        windowWidth = st.getInt( 'Settings/WindowWidth', 0 )
        windowHeight = st.getInt( 'Settings/WindowHeight', 0 )
        if (windowWidth > 0) and (windowHeight > 0) :
            windowWidth = max( windowWidth, 100 )
            windowHeight = max( windowHeight, 100 )
            self.resize( QSize(windowWidth,windowHeight).expandedTo(self.minimumSizeHint()) )
        profiles = listProfiles()
        entries = [entry for userName,keyId,entry in profiles]
        if st.getInt('Settings/RememberKey',0) :
            entry = st.getString('Settings/SavedProfile')
            password = st.getString('Settings/SavedPassword')
            if entry and password and (entry in entries) :
                profile = loadProfile( entry, password )
                if profile :
                    self._doGoOnline( profile )

    def _sortContacts( self ) :
        self.ui.contacts.sortItems()

    def _addContactItem( self, contact ) :
        contactItem = QListWidgetItem( QIcon(':/images/user_offline.png'),
                contact.name, self.ui.contacts )
        self._sortContacts()
        self.statusChecker.check( contactItem )

    def _cancelReconnect( self ) :
        if self.reconnector :
            self.reconnector.shutdown()
            self.reconnector = None
            self._updateUI()

    @pyqtSignature( '' )
    def on_connectCancelButton_clicked( self ) :
        self._cancelReconnect()
        self.exitAfterOffline = False
        self.session.shutdown()

    @pyqtSignature( '' )
    def on_actionCreateKey_triggered( self ) :
        self._cancelReconnect()
        dlg = CreateKeyDialog( self, self.reactor )
        result = dlg.exec_()
        if result == QDialog.Rejected : return
        userName = dlg.userName
        password = dlg.password
        rsaKey = dlg.rsaKey
        keyId = dlg.keyId
        profile = createProfile( rsaKey, password, userName, keyId )
        dlg = CreateKeyDoneDialog( self, keyId )
        result = dlg.exec_()
        if result == QDialog.Rejected : return
        st = localSettings()
        if dlg.rememberKey :
            st.setInt( 'Settings/RememberKey', 1 )
            st.setString( 'Settings/SavedProfile', profile.storeEntry )
            st.setString( 'Settings/SavedPassword', password )
        else :
            st.setInt( 'Settings/RememberKey', 0 )
            st.remove( 'Settings/SavedProfile' )
            st.remove( 'Settings/SavedPassword' )
        self._updateUI()
        self._doGoOnline( profile )

    @pyqtSignature( '' )
    def on_actionGoOnline_triggered( self ) :
        self._cancelReconnect()
        dlg = GoOnlineDialog( self )
        if not dlg.profiles :
            QMessageBox.warning( self, 'No Private Keys',
                    'No Private Keys have been created. Please create your Private Key first.' )
            self.ui.actionCreateKey.trigger()
            return
        result = dlg.exec_()
        if result == QDialog.Rejected : return
        self._doGoOnline( dlg.profile )

    def _doGoOnline( self, profile ) :
        self.profile = profile
        self.session.goOnline( self.profile )
        self.reconnector = Reconnector( self.profile,
                self._onReconnectTimer, self.reactor )

    def _onReconnectTimer( self ) :
        assert self.reconnector
        if self.reconnector.timeLeft > 0 :
            self._updateUI()
            return
        self.profile = self.reconnector.profile
        self._cancelReconnect()
        self.session.goOnline( self.profile )
        self.reconnector = Reconnector( self.profile,
                self._onReconnectTimer, self.reactor )

    @pyqtSignature( '' )
    def on_actionGoOffline_triggered( self ) :
        self._cancelReconnect()
        self.session.goOffline()

    @pyqtSignature( '' )
    def on_actionKeyInfo_triggered( self ) :
        assert self.sm.current() == self.ONLINE
        if self.keyInfoDialog :
            self.keyInfoDialog.show()
            self.keyInfoDialog.activateWindow()
            return
        def onDialogClose() :
            self.keyInfoDialog = None
            self.sm.removeCallback( callbackId )
        def onStateChange() :
            self.keyInfoDialog.close()
            assert not self.keyInfoDialog
        dlg = KeyInfoDialog( self, self.profile, onDialogClose )
        dlg.show()
        self.keyInfoDialog = dlg
        callbackId = self.sm.insertCallback( onStateChange,
                src=self.ONLINE )

    @pyqtSignature( '' )
    def on_actionExit_triggered( self ) :
        self._cancelReconnect()
        if self.sm.current() == self.ONLINE :
            self.exitAfterOffline = True
            self.session.goOffline()
        else :
            self.exitAfterOffline = False
            self.appletServer.shutdown()
            self.session.shutdown()
            if self.trayIcon is not None :
                self.trayIcon.hide()
            st = localSettings()
            st.setInt( 'Settings/WindowWidth', self.width() )
            st.setInt( 'Settings/WindowHeight', self.height() )
            qApp.exit( 0 )

    @pyqtSignature( '' )
    def on_actionAddContact_triggered( self ) :
        assert self.sm.current() == self.ONLINE
        if self.addContactDialog :
            self.addContactDialog.show()
            self.addContactDialog.activateWindow()
            return
        def onDialogClose() :
            self.addContactDialog = None
            self.sm.removeCallback( callbackId )
        def onAddContact( contact ) :
            self.profile.addContact( contact )
            self._addContactItem( contact )
            saveProfileContacts( self.profile )
        def onStateChange() :
            self.addContactDialog.close()
            assert not self.addContactDialog
        dlg = AddContactDialog( self, self.reactor, self.profile,
                onAddContact, onDialogClose )
        dlg.show()
        self.addContactDialog = dlg
        callbackId = self.sm.insertCallback( onStateChange,
                src=self.ONLINE )

    @pyqtSignature( '' )
    def on_actionRefreshStatus_triggered( self ) :
        assert self.sm.current() == self.ONLINE
        for i in range(self.ui.contacts.count()) :
            item = self.ui.contacts.item( i )
            self.statusChecker.check( item )

    def _onContactsContextMenuRequested( self, pos ) :
        item = self.ui.contacts.itemAt( pos )
        if item is None : return
        curItem = self.ui.contacts.currentItem()
        if item is not curItem : return
        self.contextMenu.popup( self.ui.contacts.mapToGlobal(pos) )

    def _onContactDoubleClicked( self, item ) :
        assert self.sm.current() == self.ONLINE
        item = self.ui.contacts.currentItem()
        if item is None : return
        self.actionManager.execDefaultAction( str(item.text()) )

    def _onContextMenuAction( self, action ) :
        assert self.sm.current() == self.ONLINE
        item = self.ui.contacts.currentItem()
        if item is None : return
        self.actionManager.execAction( action, str(item.text()) )

    @pyqtSignature( '' )
    def on_actionCheckStatus_triggered( self ) :
        assert self.sm.current() == self.ONLINE
        item = self.ui.contacts.currentItem()
        if item is None : return
        self.statusChecker.check( item )

    @pyqtSignature( '' )
    def on_actionRemoveContact_triggered( self ) :
        assert self.sm.current() == self.ONLINE
        item = self.ui.contacts.currentItem()
        if item is None : return
        contact = self.profile.getContactByName( str(item.text()) )
        self.ev.trigger( 'contact.remove', item, contact )
        self.ui.contacts.takeItem( self.ui.contacts.row(item) )
        self.profile.removeContact( contact )
        saveProfileContacts( self.profile )
        permissions = self.session.getPermissions()
        permissions.removeContact( contact.name )
        if permissions.isModified() :
            permissions.savePermissions()

    @pyqtSignature( '' )
    def on_actionContactInfo_triggered( self ) :
        assert self.sm.current() == self.ONLINE
        item = self.ui.contacts.currentItem()
        if item is None : return
        dlg = self.contactInfoDialogs.get( item )
        if dlg is None :
            contact = self.profile.getContactByName( str(item.text()) )
            # FIXME: why is obj needed?
            obj = Dummy()
            def onUpdateName( newName ) :
                oldName = contact.name
                self.profile.changeContactName( oldName, newName )
                saveProfileContacts( self.profile )
                item.setText( newName )
                self._sortContacts()
                permissions = self.session.getPermissions()
                permissions.changeContactName( oldName, newName )
                if permissions.isModified() :
                    permissions.savePermissions()
            def onClose() :
                if obj.callbackId is not None :
                    self.ev.remove( obj.callbackId )
                del self.contactInfoDialogs[item]
            def onContactRemoved( removedItem, contact ) :
                if removedItem is item :
                    dlg.close()
            dlg = ContactInfoDialog( self, contact, self.profile,
                    onUpdateName, onClose )
            self.contactInfoDialogs[item] = dlg
            obj.callbackId = self.ev.register( 'contact.remove',
                    onContactRemoved )
        dlg.show()
        dlg.activateWindow()

    @pyqtSignature( '' )
    def on_actionEditPermissions_triggered( self ) :
        assert self.sm.current() == self.ONLINE
        if self.editPermissionsDialog :
            self.editPermissionsDialog.show()
            self.editPermissionsDialog.activateWindow()
            return
        def onClose() :
            self.editPermissionsDialog = None
        dlg = PermissionsDialog( self, self.session.getPermissions(),
                onClose )
        dlg.show()
        self.editPermissionsDialog = dlg

    @pyqtSignature( '' )
    def on_actionAboutCSpace_triggered( self ) :
        buildNumber = currentBuildNumber()
        listenPortMsg = ''
        if self.sm.current() == self.ONLINE :
            listenPortMsg = 'Applet Server Port: %d\n' % \
                    self.appletServer.getListenPort()
        QMessageBox.information( self, 'About CSpace',
                'CSpace Peer-to-Peer Communication Platform\n'+
                'Build %d\n' % buildNumber +
                listenPortMsg +
                '(c) Tachyon Technologies 2006',
                QMessageBox.Ok )

    def _onTrayClicked( self, pos, button ) :
        if (self.isMinimized()) or (not self.isVisible()) :
            self.showNormal()
        self.activateWindow()

    def _onUpdateDownloaded( self, updateFileName, updateFileData ) :
        def onInstall() :
            installerFile = self.autoUpdater.saveInstaller(
                    updateFileName, updateFileData )
            global execFileAfterExit
            execFileAfterExit = installerFile
            self.ui.actionExit.trigger()
            self.notifyWindow.close()
        self.notifyWindow = UpdateNotifyWindow( self.reactor,
                onInstall )

    def closeEvent( self, ev ) :
        self.showMinimized()
        if self.trayIcon is not None :
            self.hide()
        ev.ignore()

class LogFile( object ) :
    def __init__( self, settings=None, appName="CSpace" ) :
        if settings is None:
            settings = localSettings()
        configDir = settings.getConfigDir()
        logFile = os.path.join( configDir, '%s.log'%appName )
        try :
            if os.path.getsize(logFile) >= 1024*1024 :
                os.remove( logFile )
        except OSError :
            pass
        self.f = file( logFile, 'a' )

    def write( self, s ) :
        self.f.write( s )
        self.f.flush()

    def flush( self ) :
        pass

def checkSettingsVersion() :
    s = localSettings()
    if not s.listEntries('') :
        s.setInt( 'SettingsVersion', SETTINGS_VERSION )
        return True
    version = s.getInt( 'SettingsVersion', -1 )
    if version == SETTINGS_VERSION : return True
    result = QMessageBox.information( None, 'Settings Update',
            'CSpace Settings format has changed. Old settings will be removed.\n' +
            'Press OK to remove data and continue.\n' +
            'Press CANCEL to exit CSpace.\n' +
            'Remove data in %s?' % s.getConfigDir(),
            QMessageBox.Ok,
            QMessageBox.Cancel | QMessageBox.Escape )
    if result != QMessageBox.Ok : return False
    for entry in s.listEntries('') :
        result = s.remove( entry )
        if not result :
            QMessageBox.critical( None, 'Error',
                    'Unable to delete entry: %s' % entry )
            return False
    s.setInt( 'SettingsVersion', SETTINGS_VERSION )
    return True

def readAddr( arg ) :
    addr = arg.split( ':' )
    addr[1] = int(addr[1])
    addr = tuple(addr)
    return addr

def main() :
    try :
        nodeAddr = readAddr( sys.argv[1] )
    except :
        nodeAddr = ('208.78.96.185', 10001) # ('210.210.1.102',10001) 
    seedNodes = [nodeAddr]
    app = QApplication( sys.argv )
    icon = QIcon()
    for iconSize in [48,32,24,16] :
        icon.addPixmap( QPixmap(':/images/cspace%d.png'%iconSize) )
    app.setWindowIcon( icon )
    app.setQuitOnLastWindowClosed( False )
    reactor = Qt4Reactor()
    if not checkSettingsVersion() :
        return
    logFile = LogFile( localSettings() )
    sys.stdout = logFile
    sys.stderr = logFile
    logging.getLogger().addHandler( logging.StreamHandler() )
    initdelaygc( reactor )
    mainWin = MainWindow( seedNodes, reactor )
    mainWin.show()
    app.exec_()
    global execFileAfterExit
    if execFileAfterExit is not None :
        p = execFileAfterExit
        args = [p]
        startingDir = os.path.split( p )[0]
        spawnProcess( p, args, os.environ, startingDir, 0 )

if __name__ == '__main__' :
    main()



_objects = []
_timerOp = None
_reactor = None

def initdelaygc( reactor ) :
    global _reactor
    _reactor = reactor

def delaygc( obj ) :
    global _timerOp
    _objects.append( obj )
    if _timerOp is not None :
        _timerOp.cancel()
    _timerOp = _reactor.callLater( 1, _onTimer )

def _onTimer() :
    global _timerOp
    _timerOp = None
    del _objects[:]



#!/usr/bin/python
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
import time
import datetime

#-------------------------------------------------------------------------------

def interval_to_next_hour():
    _struct_time = list(time.localtime())
    _struct_time[4] = 0
    _struct_time[5] = 0
    prev_hour_time = time.mktime(_struct_time)
    return prev_hour_time + 60*60 - time.time()

#-------------------------------------------------------------------------------

def shedule_continuously(last_time, interval):
    try:
        dt = time.time() - float(last_time)
        n = int(dt / float(interval))
    except:
        return None
    return float(last_time) + (n+1)*float(interval)

#-------------------------------------------------------------------------------

def shedule_next_hourly(last_time, interval):
    try:
        t = list(time.localtime(float(last_time)))
        t[4] = 0
        t[5] = 0
        while True:
            t[3] += int(interval)
            tm = time.mktime(t)
            if tm > time.time():
                return tm
    except:
        pass
    return None

#------------------------------------------------------------------------------

def shedule_next_daily(last_time, period_string, start_time_string):
    try:
        start_time_structtime = list(time.strptime(start_time_string, '%H:%M:%S'))
    except:
        try:
            start_time_structtime = list(time.strptime(start_time_string, '%H:%M'))
        except:
            import dhnio
            dhnio.DprintException()
            return None
    try:
        last_datetime = datetime.datetime.fromtimestamp(float(last_time))
        period = int(period_string)
    except:
        print 'DEBUG: next_daily1'
        import dhnio
        dhnio.DprintException()
        return None
    if period == 0:
        print 'DEBUG: next_daily2'
        return None

    start_time = datetime.time(
        hour = start_time_structtime[3], 
        minute = start_time_structtime[4], 
        second = start_time_structtime[5])

    today = datetime.datetime.today()
    time_ok = today.time() < start_time
    day_offset = today.toordinal() - last_datetime.toordinal()
    period_ok = day_offset % period == 0

    if time_ok and period_ok:
        today = today.replace(
            hour = start_time.hour, 
            minute = start_time.minute, 
            second = start_time.second, 
            microsecond = 0)
        return time.mktime(today.timetuple())

    today = today.replace(
        hour = start_time.hour, 
        minute = start_time.minute, 
        second = start_time.second, 
        microsecond = 0)
    delta_days = period - day_offset
    while delta_days < 0:
        delta_days += 365
    today = today + datetime.timedelta(days = delta_days)
    return time.mktime(today.timetuple())

#------------------------------------------------------------------------------ 

def shedule_next_weekly(last_time, period_string, start_time_string, week_days):
    try:
        start_time_structtime = list(time.strptime(start_time_string, '%H:%M:%S'))
    except:
        try:
            start_time_structtime = list(time.strptime(start_time_string, '%H:%M'))
        except:
            import dhnio
            dhnio.DprintException()
            return None
    try:
        last_datetime = datetime.datetime.fromtimestamp(float(last_time))
        period = int(period_string)
    except:
        import dhnio
        dhnio.DprintException()
        return None
    if len(week_days) == 0 or period == 0:
        print 'DEBUG: next_weekly2'
        return None

    start_time = datetime.time(
        hour = start_time_structtime[3], 
        minute = start_time_structtime[4],
        second = start_time_structtime[5])

    today = datetime.datetime.today()

    week_ok = today.weekday() in week_days
    time_ok = today.time() < start_time

    last_iso_year, last_iso_week, last_iso_weekday = last_datetime.isocalendar()
    today_iso_year, today_iso_week, today_iso_weekday = today.isocalendar()
    period_ok = (today_iso_week - last_iso_week) % period == 0

    if time_ok and week_ok and period_ok:
        today = today.replace(
            hour = start_time.hour, 
            minute = start_time.minute, 
            second = start_time.second, 
            microsecond = 0)
        return time.mktime(today.timetuple())

    today = today.replace(hour=0, minute=0, second=0, microsecond=0)

    while True:
        today = today + datetime.timedelta(days=1)
        week_ok = today.weekday() in week_days

        today_iso_year, today_iso_week, today_iso_weekday = today.isocalendar()
        period_ok = (today_iso_week - last_iso_week) % period == 0

        if week_ok and period_ok:
            break

    today = today.replace(
        hour = start_time.hour, 
        minute = start_time.minute, 
        second = start_time.second, 
        microsecond = 0)
    return time.mktime(today.timetuple())

#------------------------------------------------------------------------------ 

def shedule_next_monthly_old(last_time, day_string, start_time_string, months):
    try:
        start_time_structtime = list(time.strptime(start_time_string, '%H:%M:%S'))
    except:
        try:
            start_time_structtime = list(time.strptime(start_time_string, '%H:%M'))
        except:
            import dhnio
            dhnio.DprintException()
            return None
    try:
        last_datetime = datetime.datetime.fromtimestamp(float(last_time))
        day = int(day_string)
    except:
        import dhnio
        dhnio.DprintException()
        return None
    if len(months) == 0 or day > 31 or day < 1:
        print 'DEBUG: next_monthly2'
        return None

    start_time = datetime.time(
        hour = start_time_structtime[3], 
        minute = start_time_structtime[4],
        second = start_time_structtime[5])

    today = datetime.datetime.today()

    month_ok = today.month in months
    day_ok = today.day == day
    time_ok = today.time() < start_time

    if month_ok and day_ok and time_ok:
        today = today.replace(
            hour = start_time.hour, 
            minute = start_time.minute, 
            second = start_time.second, 
            microsecond = 0)
        return time.mktime(today.timetuple())

    today = today.replace(hour=0, minute=0, second=0, microsecond=0)

    month = today.month
    if today.day > day or not time_ok:
        month = month + 1
    today = today.replace(day=day)

    while True:
        if month > 12:
            month = 1
            today = today.replace(year=today.year+1)
        today = today.replace(month=month)
        if month in months:
            break

        month += 1

    today = today.replace(
        hour = start_time.hour, 
        minute = start_time.minute, 
        second = start_time.second, 
        microsecond = 0)
    return time.mktime(today.timetuple())

#------------------------------------------------------------------------------ 

def shedule_next_monthly(last_time, interval_months_string, start_time_string, dates):
    try:
        start_time_structtime = list(time.strptime(start_time_string, '%H:%M:%S'))
    except:
        try:
            start_time_structtime = list(time.strptime(start_time_string, '%H:%M'))
        except:
            import dhnio
            dhnio.DprintException()
            return None
    try:
        last_datetime = datetime.datetime.fromtimestamp(float(last_time))
        interval_months = int(interval_months_string)
    except:
        import dhnio
        dhnio.DprintException()
        return None

    good_dates = []
    for d in dates:
        try:
            int(d)
        except:
            continue
        good_dates.append(int(d))
            
    if len(good_dates) == 0:
        good_dates.append(1)
    
    dates = good_dates

    months = []
    for month in range(12):
        if month % interval_months == 0:
            months.append(month + 1)
            
    if len(months) == 0:
        months.append(1)
        
    start_time = datetime.time(
        hour = start_time_structtime[3], 
        minute = start_time_structtime[4],
        second = start_time_structtime[5])

    today = datetime.datetime.today()
    
    day_ok = today.day in dates
    time_ok = today.time() < start_time
    month_ok = today.month in months

    if month_ok and day_ok and time_ok:
        del months 
        today = today.replace(
            hour = start_time.hour, 
            minute = start_time.minute, 
            second = start_time.second, 
            microsecond = 0)
        return time.mktime(today.timetuple())

    day = today.day
    month = today.month

    if not time_ok:
        day += 1
        #today = today.replace(hour=0, minute=0, second=0, microsecond=0)

    if not day in dates:
        while True:
            day += 1
            if day > 31:
                day = 1
                month += 1
                if month > 12:
                    month = 1
            if month in months and day in dates:
                break

    del months
    today = today.replace(
        month = month, 
        day = day,
        hour = start_time.hour, 
        minute = start_time.minute, 
        second = start_time.second, 
        microsecond = 0)
    return time.mktime(today.timetuple())

#-------------------------------------------------------------------------------

def toInt(s, default=0):
    try:
        return int(s)
    except:
        return default

def toFloat(s, default=0.0):
    try:
        return float(s)
    except:
        return default

#------------------------------------------------------------------------------ 

#tests
if __name__ == '__main__':

#    print interval_to_next_hour()
#    t = 1286897649.59
##    print time.asctime(time.localtime(shedule_non_stop(t, 60*60)))
##    print shedule_non_stop(t, 60*60) - time.time()
#    print shedule_next_hourly(time.time()-60*60*9-12, 1)/60
    t = shedule_next_monthly(time.time()-60*60*9-12, '1', '14:00', ['2','6','9', '15', '16'])
    print time.time(), t , time.time() - t
    print time.asctime(time.localtime(t))
    print time.asctime(time.localtime())
